from collections import deque
import numpy as np
import tomopy
import pyqtgraph as pg
from PySide import QtGui, QtCore
from collections import OrderedDict
from .loader import ProjectionStack, SinogramStack
from pipeline.loader import StackImage
from pipeline import msg
from xicam.plugins.tomography import functionwidgets, functionmanager, reconpkg, config
from xicam.widgets.customwidgets import DataTreeWidget, ImageView, histDialogButton
from xicam.widgets.roiwidgets import ROImageOverlay
from xicam.widgets.imageviewers import StackViewer, ArrayViewer
from xicam.widgets.volumeviewers import VolumeViewer


__author__ = "Luis Barroso-Luque, Holden Parks"
__copyright__ = "Copyright 2016, CAMERA, LBL, ALS"
__credits__ = ["Ronald J Pandolfi", "Dinesh Kumar", "Singanallur Venkatakrishnan", "Luis Luque", "Alexander Hexemer"]
__license__ = ""
__version__ = "1.2.1"
__maintainer__ = "Ronald J Pandolfi"
__email__ = "ronpandolfi@lbl.gov"
__status__ = "Beta"


class TomoViewer(QtGui.QWidget):
    """
    Class that holds projection, sinogram, recon preview, and process-settings viewers for a tomography dataset.

    Attributes
    ----------
    data : pipeline.loader.StackImage
        Raw tomography data as a StackImage
    viewstack : QtGui.StackedWidget
        Container for different tomography viewers
    viewmode : QtGui.QTabBar
        Tabbar to switch between tomopgraphy views
    projectionViewer : ProejctionViewer
        Viewer class to visualize raw tomography projections
    sinogramViewer : StackViewer
        Viewer class to visualize raw tomography sinograms
    previewViewer : PreviewViewer
        Viewer class to hold a set of preview reconstructions of a single sinogram/slice
    preview3DViewer : Preview3DViewer
        Viewer class to visualize a reconstruction of subsampled set of the raw data
    pipeline : OrderedDict
        Dictionary to hold parameters for reconstruction, referenced by the iterations
        of the reconstruction function
    prange : list
        List of values used in TestParameterRange previews
    path : str
        Path to dataset being represented
    toolbar : ui.Toolbar
        Tomography plugin's toolbar
    preview_holder : list
        Holds slice preview data, including parameters and tomography data


    Signals
    -------
    sigSetDefaults(dict)
        emits dictionary with dataset specific defaults to set in tomography plugin

    Parameters
    ----------
    paths : str/list of str, optional
        Path to input dataset
    data : ndarry, optional
        Array with input data. Currently only paths are supported.
    args
        Additional arguments
    kwargs
        Additional keyword arguments
    """

    sigSetDefaults = QtCore.Signal(dict)

    def __init__(self, paths=None, toolbar=None, data=None, *args, **kwargs):
        if paths is None and data is None:
            raise ValueError('Either data or path to file must be provided')

        super(TomoViewer, self).__init__(*args, **kwargs)

        # pipeline dictionary of parameters
        self.pipeline = OrderedDict()

        # set path as field of TomoViewer
        self.path = paths
        self.toolbar = toolbar

        # self._recon_path = None
        self.viewstack = QtGui.QStackedWidget(self)
        self.viewmode = QtGui.QTabBar(self)
        self.viewmode.addTab('Projection View')  # TODO: Add icons!
        self.viewmode.addTab('Sinogram View')
        self.viewmode.addTab('Flats')
        self.viewmode.addTab('Darks')
        self.viewmode.addTab('Slice Preview')
        self.viewmode.addTab('3D Preview')
        self.viewmode.setShape(QtGui.QTabBar.TriangularSouth)


        # keep a timer for reconstruction
        self.recon_start_time = 0
        self.preview_holder = []
        self.prange = []

        if data is not None:
            self.data = data
        elif paths is not None and len(paths):
            if paths.endswith('.tif') or paths.endswith('.tiff'):
                self.data = self.loaddata(paths, raw=True)
            else:
                self.data = self.loaddata(paths)

        self.projectionViewer = ProjectionViewer(self.data, self.toolbar, parent=self)
        self.projectionViewer.centerBox.setRange(0, self.data.shape[1])
        self.projectionViewer.stackViewer.connectImageToName(self.data.fabimage.frames)
        self.viewstack.addWidget(self.projectionViewer)

        self.sinogramViewer = StackViewer(SinogramStack.cast(self.data), parent=self)
        self.sinogramViewer.setIndex(self.sinogramViewer.data.shape[0] // 2)
        self.viewstack.addWidget(self.sinogramViewer)

        self.flatViewer = ArrayViewer(self.data.flats,flipAxes=True,  parent=self)
        self.darkViewer = ArrayViewer(self.data.darks, flipAxes=True, parent=self)
        self.viewstack.addWidget(self.flatViewer)
        self.viewstack.addWidget(self.darkViewer)

        self.previewViewer = PreviewViewer(self.data.shape[1], parent=self)
        self.previewViewer.sigSetDefaults.connect(self.sigSetDefaults.emit)
        self.viewstack.addWidget(self.previewViewer)

        self.preview3DViewer = Preview3DViewer(parent=self)
        self.preview3DViewer.volumeviewer.moveGradientTick(1, 0.3)
        self.preview3DViewer.sigSetDefaults.connect(self.sigSetDefaults.emit)
        self.viewstack.addWidget(self.preview3DViewer)


        v = QtGui.QVBoxLayout(self)
        v.setContentsMargins(0, 0, 0, 0)
        v.addWidget(self.viewstack)
        v.addWidget(self.viewmode)
        self.setLayout(v)


        self.viewmode.currentChanged.connect(self.viewstack.setCurrentIndex)
        self.viewstack.currentChanged.connect(self.viewmode.setCurrentIndex)

    def wireupCenterSelection(self, recon_function):
        """
        Connect the reconstruction functions parameters to the manual center selection button.
        And connect the parameters sigValueChanged to the center detection image overlay widget

        Parameters
        ----------
        recon_function : FuncionWidget
            Reconstruction function widget with a 'center' child parameter

        """
        if recon_function is not None:
            center_param = recon_function.params.child('center')
            # Uncomment this if you want convenience of having the center parameter in pipeline connected to the
            # manual center widget, but this limits the center options to a resolution of 0.5
            self.projectionViewer.sigCenterChanged.connect(
                lambda x: center_param.setValue(x)) #, blockSignal=center_param.sigValueChanged))
            self.projectionViewer.centerBox.valueChanged.connect(lambda x: center_param.setValue(x))
            self.projectionViewer.setCenterButton.clicked.connect(
                lambda: center_param.setValue(self.projectionViewer.centerBox.value()))
            center_param.sigValueChanged.connect(lambda p,v: self.projectionViewer.centerBox.setValue(v))
            center_param.sigValueChanged.connect(lambda p,v: self.projectionViewer.updateROIFromCenter(v))

    def openFlats(self):
        """
        Opens new 'flats' for dataset from path taken from user via QtGui.QFileDialog
        """

        flat_dialog = QtGui.QFileDialog(self).getOpenFileName(caption="Please select flats for this dataset: ")
        path = flat_dialog[0]

        if path:
            import fabio
            msg.showMessage('Loading flats...')
            try:
                flats = fabio.open(path)
                self.data.flats = OrderedDict()
                for frame in sorted(flats.frames):
                    self.data.flats[frame] = np.squeeze(np.copy(flats._dgroup[frame])).transpose()
                self.flatViewer.setData(self.data.flats)
                del flats
            except Exception as e:
                QtGui.QMessageBox.warning(self, 'Warning', 'Flats not loaded.')
                msg.showMessage('Unable to load flats. Check log for details.', timeout=10)
                raise e
        msg.clearMessage()

    def openDarks(self):
        """
        Opens new 'darks' for dataset from path taken from user via QtGui.QFileDialog
        """
        dark_dialog = QtGui.QFileDialog(self).getOpenFileName(caption="Please select darks for this dataset: ")
        path = dark_dialog[0]

        if path:
            import fabio
            msg.showMessage('Loading darks...', timeout=10)
            try:
                darks = fabio.open(path)
                self.data.darks = OrderedDict()
                for frame in sorted(darks.frames):
                    self.data.darks[frame] = np.squeeze(np.copy(darks._dgroup[frame])).transpose()
                self.darkViewer.setData(self.data.darks)
                del darks
            except Exception as e:
                QtGui.QMessageBox.warning(self, 'Warning', 'Darks not loaded.')
                msg.showMessage('Unable to load darks. Check log for details.', timeout=10)
                raise e
        msg.clearMessage()

    @staticmethod
    def loaddata(paths, raw=True):
        """
        Load data from a file or list of files

        Parameters
        ----------
        paths : str/list of str
            Path to files
        raw : bool
            Boolean specifiying it the file is a raw dataset with flats and darks
            (not using this now but can be used for files where flats/darks are in seperate files)

        Returns
        -------
        ProjectionStack, StackImage:
            Class with raw data from file

        """

        if raw:
            return ProjectionStack(paths)
        else:
            return StackImage(paths)


    def getsino(self, slc=None): #might need to redo the flipping and turning to get this in the right orientation
        """
        Returns the sinograms specified in slc (this and getproj can be made one function)

        Parameters
        ----------
        slc : slice, optional
            Slice object specifying the portion of the array to return

        Returns
        -------
        ndarray:
            Array of raw data

        """
        if slc is None:
            return np.ascontiguousarray(self.sinogramViewer.currentdata[:, np.newaxis, :])
        else:
            return np.ascontiguousarray(self.data.fabimage[slc])

    def getproj(self, slc=None):
        """
        Returns the projections specified in slc (this and getsino can be made one function)

        Parameters
        ----------
        slc : slice, optional
            Slice object specifying the portion of the array to return

        Returns
        -------
        ndarray:
            Array of raw data

        """
        if slc is None:
            return np.ascontiguousarray(self.projectionViewer.currentdata[np.newaxis, :, :])
        else:
            return np.ascontiguousarray(self.data.fabimage[slc])

    def getflats(self, slc=None):
        """
        Returns the flat fields specified in slc

        Parameters
        ----------
        slc : slice, optional
            Slice object specifying the portion of the array to return

        Returns
        -------
        ndarray:
            Array of flat field data

        """
        flats = self.data.flats
        if hasattr(flats,'values'): flats = np.array(flats.values())
        if slc is None:
            return np.ascontiguousarray(flats[:, self.sinogramViewer.currentIndex, :])
        else:
            return np.ascontiguousarray(flats[slc])

    def getdarks(self, slc=None):
        """
        Returns the dark fields specified in slc

        Parameters
        ----------
        slc : slice, optional
            Slice object specifying the portion of the array to return

        Returns
        -------
        ndarray:
            Array of dark field data

        """
        darks = self.data.darks
        if hasattr(darks, 'values'): darks = np.array(darks.values())
        if slc is None:
            return np.ascontiguousarray(darks[:, self.sinogramViewer.currentIndex, :])
        else:
            return np.ascontiguousarray(darks[slc])

    def getheader(self):
        """Return the data's header (metadata)"""
        return self.data.header

    def addSlicePreview(self, params, recon, slice_no=None, prange=None):
        """
        Adds a slice reconstruction preview with the corresponding workflow pipeline dictionary to the previewViewer

        Parameters
        ----------
        params : dict
            Pipeline dictionary
        recon : ndarry
            Reconstructed slice
        slice_no : int, optional
            Sinogram/slice number reconstructed
        prange : dict, optional
            Dictionary of parameter being tested in TestParameterRange, and the functino it belongs to
        """
        if type(recon) == str:
            return

        if slice_no is None:
            slice_num = self.sinogramViewer.view_spinBox.value()
            self.previewViewer.addPreview(np.rot90(recon[0],1), params, slice_num)
        elif type(slice_no) is list:
            for item in range(slice_no[1]- slice_no[0]+1):
                self.previewViewer.addPreview(np.rot90(recon[item], 1), params, item+slice_no[0])
        # this block ensures that the previews are added in order if testparamrange is triggered
        elif prange:
            dummy_prange = dict(prange)
            func = dummy_prange.pop('function')
            param = dummy_prange.keys()[0]

            if len(self.prange) < 1 and recon is not None:
                self.prange = prange[param]

            # this if loop and try statement ensure no errors due to the recursion below
            if recon is not None:
                self.preview_holder.append([recon, params, slice_no])
            try:
                top_val = self.prange[0]
            except IndexError:
                pass

            # run through each recon in the preview_holder, and add them to the preview viewer if the top param in
            # prange matches the preview metadata
            for index, rec in enumerate(self.preview_holder):
                for key in rec[1].iterkeys():
                    if func in key:
                        subfunc = rec[1][key].keys()[0]
                        param_val = rec[1][key][subfunc][param]
                if top_val == param_val:
                    self.previewViewer.addPreview(np.rot90(rec[0][0], 1), rec[1], rec[2])
                    self.preview_holder.pop(index)
                    self.prange = np.delete(self.prange, 0)
                    self.addSlicePreview(params, None, slice_no, prange=prange)

        else:
            self.previewViewer.addPreview(np.rot90(recon[0],1), params, slice_no)
        self.viewstack.setCurrentWidget(self.previewViewer)
        msg.clearMessage()

    def add3DPreview(self, params, recon):
        """
        Adds a slice reconstruction preview with the corresponding workflow pipeline dictionary to the preview3DViewer

        Parameters
        ----------
        params : dict
            Pipeline dictionary
        recon : ndarry
            Reconstructed array

        """
        if type(recon) == str:
            return

        recon = np.flipud(recon)
        self.viewstack.setCurrentWidget(self.preview3DViewer)
        self.preview3DViewer.setPreview(recon, params)
        hist = self.preview3DViewer.volumeviewer.getHistogram()

        # disable auto scale
        preview_hist = self.preview3DViewer.volumeviewer.HistogramLUTWidget
        preview_hist.vb.enableAutoRange(preview_hist.vb.YAxis, False)

        max = hist[0][np.argmax(hist[1])]
        self.preview3DViewer.volumeviewer.setLevels([max, hist[0][-1]])

    def onManualCenter(self, active):
        """
        Activates the manual center portion of the ProjectionViewer.
        This is connected to the corresponding toolbar signal

        Parameters
        ----------
        active : bool
            Boolean specifying to activate or not. True activate, False deactivate

        """

        if active:
            self.viewstack.setCurrentWidget(self.projectionViewer)
            self.projectionViewer.hideMBIR()
            self.projectionViewer.showCenterDetection()

        else:
            self.projectionViewer.hideCenterDetection()

    def onMBIR(self, active):
        """
        Slot to activate MBIR slurm generation menu. Not currently in use.
        """

        if active:
            self.viewstack.setCurrentWidget(self.projectionViewer)
            self.projectionViewer.hideCenterDetection()
            self.projectionViewer.showMBIR()
        else:
            self.projectionViewer.hideMBIR()

    def onROIselection(self):
        """
        Shows a rectangular roi to select portion of data to reconstruct. (Not implemented yet)

        Parameters
        ----------
        active : bool
            Boolean specifying to activate or not. True activate, False deactivate

        """
        self.viewstack.setCurrentWidget(self.projectionViewer)
        self.projectionViewer.addROIselection()




class MBIRViewer(QtGui.QWidget):
    """
    Widget to generate .slurm files used in NERSC-based MBIR
    """


    def __init__(self, data, path, *args, **kwargs):
        super(MBIRViewer, self).__init__(*args, **kwargs)
        self.mdata = data.header
        if path is list:
            paths = path[0]
        else:
            paths = path
        self.path = paths
        self.data = data
        self.center = 0
        self.cor_detection_funcs = ['Phase Correlation', 'Vo', 'Nelder-Mead']

        self.runButton = QtGui.QPushButton(parent=self)
        self.runButton.setSizePolicy(QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed, QtGui.QSizePolicy.Fixed))
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("gui/icons_34.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.runButton.setIcon(icon)
        # self.runButton.setToolTip("Submit MBIR job to NERSC")
        self.runButton.setToolTip("Generate slurm file")

        self.cor_widget = QtGui.QWidget() #parent widget for center of rotation input

        # set up widget for user choice of manual or auto COR detection
        self.cor_Holder = QtGui.QGroupBox('Center of Rotation', parent = self.cor_widget)
        manual_cor = QtGui.QRadioButton('Manually input center of rotation')
        manual_cor.clicked.connect(self.manualCOR)
        auto_cor = QtGui.QRadioButton('Auto-detect center of rotation')
        auto_cor.clicked.connect(self.autoCOR)
        manual_cor.setChecked(True)
        vbox = QtGui.QVBoxLayout()
        vbox.addWidget(manual_cor)
        vbox.addWidget(auto_cor)
        self.cor_Holder.setLayout(vbox)

        # series of widgets for manual COR input
        self.cor_Value = QtGui.QStackedWidget(parent = self.cor_widget)

        self.manual_tab = QtGui.QWidget()
        self.val_box = QtGui.QDoubleSpinBox(parent = self.manual_tab)
        self.val_box.setRange(0,10000)
        self.val_box.setDecimals(1)
        self.val_box.setValue(int(data.shape[1])/2)
        text_label = QtGui.QLabel('Center of Rotation: ', parent = self.manual_tab)
        text_layout = QtGui.QHBoxLayout()
        text_layout.addWidget(text_label)
        text_layout.addWidget(self.val_box)
        self.manual_tab.setLayout(text_layout)

        self.auto_tab = QtGui.QWidget()
        self.auto_tab_layout = QtGui.QVBoxLayout()

        self.cor_function = functionwidgets.FunctionWidget(name="Center Detection", subname="Phase Correlation",
                                package=reconpkg.packages[config.names["Phase Correlation"][1]])
        self.cor_params = pg.parametertree.Parameter.create(name=self.cor_function.name,
                                             children=config.parameters[self.cor_function.subfunc_name], type='group')
        self.cor_param_tree = pg.parametertree.ParameterTree()
        self.cor_param_tree.setMinimumHeight(200)
        self.cor_param_tree.setMinimumWidth(200)
        self.cor_param_tree.setParameters(self.cor_params,showTop = False)
        for key, val in self.cor_function.param_dict.iteritems():
            if key in [p.name() for p in self.cor_params.children()]:
                self.cor_params.child(key).setValue(val)
                self.cor_params.child(key).setDefault(val)

        self.cor_method_box = QtGui.QComboBox()
        self.cor_method_box.currentIndexChanged.connect(self.changeCORfunction)
        for item in self.cor_detection_funcs:
            self.cor_method_box.addItem(item)
        cor_method_label = QtGui.QLabel('COR detection function: ')
        cor_method_layout = QtGui.QHBoxLayout()
        cor_method_layout.addWidget(cor_method_label)
        cor_method_layout.addWidget(self.cor_method_box)

        self.auto_tab_layout.addLayout(cor_method_layout)
        self.auto_tab_layout.addWidget(self.cor_param_tree)
        self.auto_tab.setLayout(self.auto_tab_layout)

        # set up COR stackwidget
        self.cor_Value.addWidget(self.manual_tab)
        self.cor_Value.addWidget(self.auto_tab)

        # set up COR widget
        v = QtGui.QVBoxLayout()
        v.addWidget(self.cor_Holder)
        v.addWidget(self.cor_Value)
        self.cor_widget.setLayout(v)


        self.runButton.clicked.connect(self.write_slurm)

        self.mbirParams = pg.parametertree.ParameterTree()
        self.mbirParams.setMinimumHeight(230)
        params = [{'name': 'Dataset path', 'type': 'str'},
                  {'name': 'Z start', 'type': 'int', 'value': 0, 'default': 0},
                  {'name': 'Z num elts', 'type': 'int', 'value': int(data.shape[-1]) ,
                   'default': int(data.shape[-1])},
                  {'name': 'Smoothness', 'type': 'float', 'value': 0.15, 'default': 0.15},
                  {'name': 'Zinger thresh', 'type': 'float', 'value': 5, 'default': 5},
                  {'name': 'View subsample factor', 'type': 'int', 'value': 2, 'default': 2},
                  {'name': 'Output folder', 'type':'str', 'value':'Results', 'default': 'Results'}]

        self.mbir_params = pg.parametertree.Parameter.create(name='MBIR Parameters', type='group', children=params)
        self.mbirParams.setParameters(self.mbir_params,showTop=False)


        right_menu = QtGui.QSplitter(self)
        right_menu.setOrientation(QtCore.Qt.Vertical)
        button_holder = QtGui.QStackedWidget()
        button_holder.addWidget(self.runButton)
        right_menu.addWidget(button_holder)
        right_menu.addWidget(self.cor_widget)

        container = QtGui.QWidget()
        container_layout = QtGui.QVBoxLayout()
        container_layout.addWidget(right_menu)
        container.setLayout(container_layout)

        left_menu = QtGui.QSplitter(self)
        left_menu.addWidget(self.mbirParams)
        left_menu.addWidget(container)


        h = QtGui.QHBoxLayout()
        h.setContentsMargins(0, 0, 0, 0)
        h.addWidget(left_menu)


        self.setLayout(h)

    def changeCORfunction(self, index):
        """
        Changes COR auto-detect function used based on index of combobox
        """

        subname = self.cor_method_box.itemText(index)
        self.auto_tab_layout.removeWidget(self.cor_param_tree)

        self.cor_function = functionwidgets.FunctionWidget(name="Center Detection", subname=subname,
                                package=reconpkg.packages[config.names[subname][1]])
        self.cor_params = pg.parametertree.Parameter.create(name=self.cor_function.name,
                                             children=config.parameters[self.cor_function.subfunc_name], type='group')
        self.cor_param_tree = pg.parametertree.ParameterTree()
        self.cor_param_tree.setMinimumHeight(200)
        self.cor_param_tree.setMinimumWidth(200)
        self.cor_param_tree.setParameters(self.cor_params,showTop = False)
        for key, val in self.cor_function.param_dict.iteritems():
            if key in [p.name() for p in self.cor_params.children()]:
                self.cor_params.child(key).setValue(val)
                self.cor_params.child(key).setDefault(val)

        self.auto_tab_layout.addWidget(self.cor_param_tree)
        self.auto_tab.setLayout(self.auto_tab_layout)



    def manualCOR(self):
        """Slot to receive signal when manual COR detection is chosen"""
        self.cor_Value.setCurrentWidget(self.manual_tab)

    def autoCOR(self):
        """Slot to receive signal when automatic COR detection is chosen"""
        self.cor_Value.setCurrentWidget(self.auto_tab)

    def loadCOR(self):
        """
        Get dataset COR, either automatically or from user-entered value
        :return:
        """

        widget = self.cor_Value.currentWidget()
        if widget is self.manual_tab:
            return self.val_box.value()
        else:
            if self.parentWidget():
                return self.find_COR(self.cor_function.subfunc_name)
            else:
                return -1

    def find_COR(self, cor_function):
        """Auto-detect COr based on 'cor_function' parameter"""

        if not cor_function in self.cor_detection_funcs:
            return -1
        else:
            if cor_function == 'Phase Correlation':
                proj1, proj2 = map(self.data.fabimage.__getitem__, (0,-1))
                kwargs = {'proj1' : proj1, 'proj2' : proj2}
            elif cor_function == 'Vo':
                kwargs = {'tomo' : np.ascontiguousarray(self.data.fabimage[:, :, :])}
            elif cor_function == 'Nelder-Mead':
                kwargs = {'tomo' : np.ascontiguousarray(self.data.fabimage[:, :, :]),
                          'theta' : tomopy.angles(int(self.data.shape[0]),ang1=90,ang2=270)}
            else:
                return -1

            for child in self.cor_params.children():
                kwargs[child.name()] = child.value()

            val = self.cor_function._function(**kwargs)
            return val[0] if val is list else val



    def write_slurm(self):
        """
        A 'slurm' file is a job to run on nersc
        """

        import os.path
        msg.showMessage("Generating slurm file...", timeout=0)

        self.center = self.loadCOR()

        if not self.center > 0 or self.center > self.data.shape[0]:
            msg.showMessage('Invalid center of rotation')
            pass
        else:

            views = int(self.data[0]) - 1
            file_name = self.path.split("/")[-1].split(".")[0]
            nodes = int(np.ceil(self.mbir_params.child('Z num elts').value()/ float(24)))
            output = os.path.join('/',self.mbir_params.child('Output folder').value(), file_name + '_mbir')

            try:
                group = self.mdata['archdir'].split("\\")[-1]
                px_size = float(self.mdata['pzdist'])*1000
            except KeyError:
                msg.showMessage('Insufficient metadata to write slurm file.', timeout=0)
            if group != file_name:
                group_hdf5 = "{}/{}".format(group, file_name)
            else:
                group_hdf5 = file_name

            slurm = '#!/bin/tcsh\n#SBATCH -p regular\n#SBATCH -N {}\n'.format(nodes)
            slurm += '#SBATCH -t 4:00:00\n#SBATCH -J {}\n#SBATCH -e {}.err\n#SBATCH -o {}.out\n\n'.format(file_name, file_name, file_name)
            slurm += 'setenv OMP_NUM_THREADS 24\nsetenv CRAY_ROOTFS DSL\nmodule load PrgEnv-intel\n'
            slurm += 'module load python/2.7.3\nmodule load h5py\nmodule load pil\nmodule load mpi4py\n\n'
            slurm += 'mkdir $SCRATCH/LaunchFolder\nmkdir $SCRATCH/Results\n\n'
            slurm += 'python XT_MBIR_3D.py --setup_launch_folder --run_reconstruction --Edison'
            slurm += ' --input_hdf5 {}/{}.h5'.format(self.mbir_params.child('Dataset path').value(), file_name)
            slurm += ' --group_hdf5 /{}'.format(group_hdf5)
            slurm += ' --code_launch_folder $SCRATCH/LaunchFolder/'
            slurm += ' --output_hdf5 $SCRATCH/Results/{}_mbir/ --x_width {}'.format(file_name, int(self.data.shape[0]))
            slurm += ' --recon_x_width {} --num_dark {}'.format(str(int(self.data.shape[0])), str(len(self.data.fabimage.darks)))
            slurm += ' --num_bright {} --z_numElts {}'.format(str(self.data.fabimage.flats),self.mbir_params.child('Z num elts').value())
            slurm += ' --z_start {} --num_views {}'.format(self.mbir_params.child('Z start').value(), views)
            slurm += ' --pix_size {} --rot_center {}'.format(px_size, self.center)
            slurm += ' --smoothness {} --zinger_thresh {}'.format(self.mbir_params.child('Smoothness').value(),
                                                                 self.mbir_params.child('Zinger thresh').value())
            slurm += ' --Variance_Est 1 --num_threads 24 --num_nodes {} '.format(nodes)
            slurm += '--view_subsmpl_fact {}'.format(self.mbir_params.child('View subsample factor').value())



            parent_folder = self.path.split(self.path.split('/')[-1])[0]
            write = os.path.join(parent_folder, '{}.slurm'.format(file_name))

            with open(write, 'w') as job:
                job.write(slurm)

            msg.showMessage("Done.", timeout=0)



class ProjectionViewer(QtGui.QWidget):
    """
    Class that holds a stack viewer, an ROImageOverlay and a few widgets to allow manual center detection

    Attributes
    ----------
    stackViewer : StackViewer
        widgets.StackViewer used to display the data
    data : loader.StackImage
        Image data
    imageoverlay_roi : widgets.ROIImageOverlay
        Widget used in the cor_widget for manual center detection
    selection_roi : pyqtgragh.ROI
        ROI for selecting region to reconstruct
    cor_widget : QtGui.QWidget
        Widget used in manual center detection
    setCenterButton : QtGui.QToolButton
        Button for setting center value from cor_widget to reconstruction function in pipeline
    roi_histogram : pyqtgraph.HistogramLUTWidget
        Histogram for imageoverlay
    mbir_viewer : MBIRViewer
        Menu for generating slurm files for NERSC-based MBIR jobs
    cor_box : QtGui.QStackWidget
        Widget for holding COR - related widgets
    cor_widget : QtGui.QWidget
        Widget for holding manual COR-related widgets
    auto_cor_widget : QtGui.QWidget
        Widget for holding automatic COR-related widgets

    Signals
    -------
    sigCenterChanged(float)
        emits float with new center value

    Parameters
    ----------
    data : pipeline.loader.StackImage
        Raw tomography data as a StackImage
    view_label : str
        String to show in QLabel lower right hand corner. Where the current index is displayed
    center : float
        center of rotation value
    args
        Additional arguments
    kwargs
        Additional keyword arguments
    """

    sigCenterChanged = QtCore.Signal(float)
    sigCORChanged = QtCore.Signal(bool)
    sigROIWidgetChanged = QtCore.Signal(pg.ROI)

    def __init__(self, data, toolbar=None, view_label=None, center=None, paths=None, *args, **kwargs):
        super(ProjectionViewer, self).__init__(*args, **kwargs)

        self.setMinimumHeight(200)

        self.stackViewer = StackViewer(data, view_label=view_label)
        self.toolbar = toolbar
        self.imageItem = self.stackViewer.imageItem
        self.data = self.stackViewer.data
        self.normalized = False
        self.imgoverlay_roi = ROImageOverlay(self.data, self.imageItem, [0, 0], parent=self.stackViewer.view)
        self.imageItem.sigImageChanged.connect(self.imgoverlay_roi.updateImage)
        self.stackViewer.view.addItem(self.imgoverlay_roi)
        self.roi_histogram = pg.HistogramLUTWidget(image=self.imgoverlay_roi.imageItem, parent=self.stackViewer)
        self.roi_histogram.vb.enableAutoRange(self.roi_histogram.vb.YAxis, False) #disable autoscaling for histogram
        self.mbir_viewer = MBIRViewer(self.data, path = self.parentWidget().path, parent=self)



        # roi to select region of interest
        self.selection_roi = None

        self.stackViewer.ui.gridLayout.addWidget(self.roi_histogram, 0, 3, 1, 2)
        self.stackViewer.keyPressEvent = self.keyPressEvent

        self.cor_widget = QtGui.QWidget(self)
        self.auto_cor_widget = functionwidgets.CORSelectionWidget(parent=self)
        self.cor_widget.setSizePolicy(QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Minimum)
        self.auto_cor_widget.setSizePolicy(QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Minimum)
        self.cor_widget.setMinimumHeight(50)
        self.auto_cor_widget.setMinimumHeight(50)

        self.cor_box = QtGui.QStackedWidget(self)
        self.cor_box.addWidget(self.auto_cor_widget)
        self.cor_box.addWidget(self.cor_widget)

        self.cor_button_holder = QtGui.QGroupBox(parent = self)
        h = QtGui.QHBoxLayout()
        self.manual_cor_button = QtGui.QRadioButton('Manually input center of rotation')
        self.manual_cor_button.clicked.connect(self.manualCOR)
        self.auto_cor_button = QtGui.QRadioButton('Auto-detect center of rotation')
        self.auto_cor_button.clicked.connect(self.autoCOR)
        self.auto_cor_button.setChecked(True)
        write_cor = QtGui.QPushButton('Write COR to metadata')
        write_cor.clicked.connect(self.writeCOR)
        h.addWidget(self.auto_cor_button)
        h.addWidget(self.manual_cor_button)
        h.addWidget(write_cor)
        self.cor_button_holder.setLayout(h)

        # push button for overlay widget's histogram range selection
        self.setButton = histDialogButton('Set', parent=self)
        self.setButton.connectToHistWidget(self.roi_histogram)
        self.stackViewer.ui.gridLayout.addWidget(self.setButton, 1, 3, 1, 2)

        clabel = QtGui.QLabel('Rotation Center:')
        olabel = QtGui.QLabel('Offset:')
        self.centerBox = QtGui.QDoubleSpinBox(parent=self.cor_widget) #QtGui.QLabel(parent=self.cor_widget)
        self.centerBox.setDecimals(1)
        self.setCenterButton = QtGui.QToolButton()
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_45.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.setCenterButton.setIcon(icon)
        self.setCenterButton.setToolTip('Set center in pipeline')
        originBox = QtGui.QLabel(parent=self.cor_widget)
        originBox.setText('x={}   y={}'.format(0, 0))
        center = center if center is not None else data.shape[1]/2.0
        self.centerBox.setValue(center) #setText(str(center))
        h1 = QtGui.QHBoxLayout()
        h1.setAlignment(QtCore.Qt.AlignLeft)
        h1.setContentsMargins(0, 0, 0, 0)
        h1.addWidget(clabel)
        h1.addWidget(self.centerBox)
        h1.addWidget(self.setCenterButton)
        h1.addWidget(olabel)
        h1.addWidget(originBox)

        # hide center button since cor updates automatically in pipeline
        self.setCenterButton.hide()

        plabel = QtGui.QLabel('Overlay Projection No:')
        plabel.setAlignment(QtCore.Qt.AlignRight)
        spinBox = QtGui.QSpinBox(parent=self.cor_widget)
        spinBox.setRange(0, data.shape[0]-1)
        slider = QtGui.QSlider(orientation=QtCore.Qt.Horizontal, parent=self.cor_widget)
        slider.setRange(0, data.shape[0]-1)
        spinBox.setValue(data.shape[0]-1)
        slider.setValue(data.shape[0]-1)
        flipCheckBox = QtGui.QCheckBox('Flip Overlay', parent=self.cor_widget)
        flipCheckBox.setChecked(True)
        constrainYCheckBox = QtGui.QCheckBox('Constrain Y', parent=self.cor_widget)
        constrainYCheckBox.setChecked(True)
        constrainXCheckBox = QtGui.QCheckBox('Constrain X', parent=self.cor_widget)
        constrainXCheckBox.setChecked(False)
        # rotateCheckBox = QtGui.QCheckBox('Enable Rotation', parent=self.cor_widget)
        # rotateCheckBox.setChecked(False)
        self.normCheckBox = QtGui.QCheckBox('Normalize', parent=self.cor_widget)
        h2 = QtGui.QHBoxLayout()
        h2.setAlignment(QtCore.Qt.AlignLeft)
        h2.setContentsMargins(0, 0, 0, 0)
        h2.addWidget(plabel)
        h2.addWidget(spinBox)
        h2.addWidget(flipCheckBox)
        h2.addWidget(constrainXCheckBox)
        h2.addWidget(constrainYCheckBox)
        # h2.addWidget(rotateCheckBox) # This needs to be implemented correctly
        h2.addWidget(self.normCheckBox)
        h2.addStretch(1)
        spinBox.setFixedWidth(spinBox.width())
        v2 = QtGui.QVBoxLayout(self.cor_widget)
        v2.addLayout(h1)
        v2.addLayout(h2)
        v2.addWidget(slider)

        l = QtGui.QGridLayout(self)
        l.setContentsMargins(0, 0, 0, 0)
        cor_holder = QtGui.QSplitter()
        cor_holder.setOrientation(QtCore.Qt.Vertical)
        cor_holder.addWidget(self.cor_box)
        cor_holder.addWidget(self.stackViewer)
        l.addWidget(self.cor_button_holder)
        l.addWidget(cor_holder)
        l.addWidget(self.mbir_viewer)
        self.hideMBIR()
        # self.mbir_viewer.hide()

        slider.valueChanged.connect(spinBox.setValue)
        slider.valueChanged.connect(self.stackViewer.resetImage)
        spinBox.valueChanged.connect(self.changeOverlayProj)
        flipCheckBox.stateChanged.connect(self.flipOverlayProj)
        constrainYCheckBox.stateChanged.connect(lambda v: self.imgoverlay_roi.constrainY(v))
        constrainXCheckBox.stateChanged.connect(lambda v: self.imgoverlay_roi.constrainX(v))

        # rotateCheckBox.stateChanged.connect(self.addRotateHandle)
        self.normCheckBox.stateChanged.connect(self.normalize)
        self.stackViewer.sigTimeChanged.connect(lambda: self.normalize(False))
        self.imgoverlay_roi.sigTranslated.connect(self.setCenter)
        self.imgoverlay_roi.sigTranslated.connect(lambda x, y: originBox.setText('x={}   y={}'.format(x, y)))
        self.hideCenterDetection()

        self.bounds = None
        # self.normalize(True)

    def updateCORChoice(self, boolean):
        """
        Slot to receive signal emitted when user chooses to use either automatic or manual COR detection in
        function pipeline
        """
        if self.toolbar and self.toolbar.actionCenter.isChecked():
            if boolean:
                self.cor_box.setCurrentWidget(self.auto_cor_widget)
                self.stackViewer.hide()
                self.auto_cor_button.setChecked(True)
            else:
                self.cor_box.setCurrentWidget(self.cor_widget)
                self.stackViewer.show()
                self.manual_cor_button.setChecked(True)

    def writeCOR(self):
        """
        Writes COR value acquired from user to metadata of input file
        """
        cor = QtGui.QInputDialog.getDouble(self.cor_box, 'Write COR value to file',
                                           'Write COR value to file',self.data.shape[1]/2)
        if cor[1]:
            self.data.fabimage.change_dataset_attribute('center', cor[0])


    def manualCOR(self):
        """
        Slot to receive signal when manual COR detection button is clicked in CORSelectionWidget
        """
        self.cor_box.setCurrentWidget(self.cor_widget)
        self.stackViewer.show()
        self.sigCORChanged.emit(False)

    def autoCOR(self):
        """
        Slot to receive signal when auto COR detection button is clicked in CORSelectionWidget
        """
        self.cor_box.setCurrentWidget(self.auto_cor_widget)
        self.stackViewer.hide()
        self.sigCORChanged.emit(True)



    def changeOverlayProj(self, idx):
        """
        Changes the image in the overlay. This is connected to the slider in the cor_widget
        """

        self.normCheckBox.setChecked(False)
        self.imgoverlay_roi.setCurrentImage(idx)
        self.imgoverlay_roi.updateImage()

    def setCenter(self, x, y):
        """
        Sets the center in the centerBox based on the position of the imageoverlay

        Parameters
        ----------
        x : float
            x-coordinate of overlay image in the background images coordinates
        y : float
            x-coordinate of overlay image in the background images coordinates
        """

        center = (self.data.shape[1] + x - 1)/2.0 # subtract half a pixel out of 'some' convention?
        self.centerBox.setValue(center)
        self.sigCenterChanged.emit(center)

    def hideCenterDetection(self):
        """
        Hides the center detection widget and corresponding histogram
        """
        self.normalize(False)
        self.cor_box.hide()
        self.cor_button_holder.hide()
        self.roi_histogram.hide()
        self.setButton.hide()
        self.imgoverlay_roi.setVisible(False)
        self.stackViewer.show()

    def showCenterDetection(self):
        """
        Shows the center detection widget and corresponding histogram
        """
        # self.normalize(True)
        self.cor_box.show()
        self.cor_button_holder.show()
        self.roi_histogram.show()
        self.setButton.show()
        self.imgoverlay_roi.setVisible(True)

        if self.auto_cor_button.isChecked():
            self.stackViewer.hide()
        else:
            self.stackViewer.show()

    def showMBIR(self):
        """
        Slot to receive signal and show MBIR menu when it is requested
        """

        self.mbir_viewer.show()
        self.cor_button_holder.hide()
        # self.hideCenterDetection()
        self.stackViewer.hide()

    def hideMBIR(self):
        """
        Slot to receive signal and show MBIR menu when it is requested
        """
        self.mbir_viewer.hide()
        self.stackViewer.show()


    def updateROIFromCenter(self, center):
        """
        Updates the position of the ROIImageOverlay based on the given center

        Parameters
        ----------
        center : float
            Location of center of rotation
        """

        s = self.imgoverlay_roi.pos()[0]
        self.imgoverlay_roi.translate(pg.Point((2 * center + 1 - self.data.shape[1] - s, 0))) # 1 again due to the so-called COR
                                                                                   # conventions...
    def flipOverlayProj(self, val):
        """
        Flips the image show in the ROIImageOverlay
        """

        self.imgoverlay_roi.flipCurrentImage()
        self.imgoverlay_roi.updateImage()

    def toggleRotateHandle(self, val):
        """
        Adds/ removes a handle on the ROIImageOverlay to be able to rotate the image (Rotation is not implemented
        correctly yet)

        Parameters
        ----------
        val : bool
            Boolean specifying to add or remove the handle
        """

        if val:
            self.toggleRotateHandle.handle = self.imgoverlay_roi.addRotateHandle([0, 1], [0.2, 0.2])
        else:
            self.imgoverlay_roi.removeHandle(self.toggleRotateHandle.handle)

    def addROIselection(self):
        """
        Adds/ removes a rectangular ROI to select a region of interest for reconstruction. Not implemented yet
        """
        if self.selection_roi:
            self.stackViewer.view.removeItem(self.selection_roi)

        self.selection_roi = pg.ROI([0, 0], [100, 100])
        self.stackViewer.view.addItem(self.selection_roi)
        self.selection_roi.addScaleHandle([1, 1], [0, 0])
        self.selection_roi.addScaleHandle([0, 0], [1, 1])
        self.sigROIWidgetChanged.emit(self.selection_roi)


    def normalize(self, val):
        """
        Toggles the normalization of the ROIImageOverlay.

        Parameters
        ----------
        val : bool
            Boolean specifying to normalize image
        """
        # self.roi_histogram.setLevels(0,1)
        if val and not self.normalized:
            if not hasattr(self.data, 'flats') or not hasattr(self.data, 'darks'):
                msg.showMessage('Must load flats and darks to normalize dataset', timeout=10)
                return
            flats = np.array(self.data.flats.values())
            darks = np.array(self.data.darks.values())
            self.flat = np.median(flats, axis=0).transpose()
            self.dark = np.median(darks, axis=0).transpose()

            proj = (self.imageItem.image - self.dark)/(self.flat - self.dark)
            overlay = self.imgoverlay_roi.currentImage
            if self.imgoverlay_roi.flipped:
                overlay = np.flipud(overlay)
            overlay = (overlay - self.dark)/(self.flat - self.dark)
            if self.imgoverlay_roi.flipped:
                overlay = np.flipud(overlay)
            self.imgoverlay_roi.currentImage = overlay

            self.imgoverlay_roi.updateImage(autolevels=True)
            self.stackViewer.setImage(proj, autoRange=False, autoLevels=True)
            self.stackViewer.updateImage()
            self.normalized = True
            self.roi_histogram.setLevels(-1, 1) # lazy solution, could be improved with some sampling methods
            self.roi_histogram.vb.setRange(yRange=(-1.5, 1.5))
            self.normCheckBox.setChecked(True)
        elif not val and self.normalized:
            self.stackViewer.resetImage()
            self.imgoverlay_roi.resetImage()
            min, max = self.stackViewer.quickMinMax(self.imgoverlay_roi.imageItem.image)
            self.roi_histogram.setLevels(min, max)
            self.normalized = False
            self.normCheckBox.setChecked(False)

    def keyPressEvent(self, ev):
        """
        Override QWidgets key pressed event to send the event to the ROIImageOverlay when it is pressed
        """
        super(ProjectionViewer, self).keyPressEvent(ev)
        if self.imgoverlay_roi.isVisible():
            self.imgoverlay_roi.keyPressEvent(ev)
        else:
            super(StackViewer, self.stackViewer).keyPressEvent(ev)
        ev.accept()


class PreviewViewer(QtGui.QSplitter):
    """
    Viewer class to show reconstruction previews in a PG ImageView, along with the function pipeline settings for the
    corresponding preview

    Attributes
    ----------
    previews : ArrayDeque
        ArrayDeque to hold slice reconstruction previews
    data : deque of dicts
        deque holding preview dicts corresponding to the reconstruction in previews
    datatrees : deque of DataTree widgets
        deque holding DataTree widgets to show the data in data deque
    slice_numbers : deque
        deque with sinogram index that was reconstructed for that preview
    imageview : widgets.ImageView
        ImageView to display preview reconstructions

    Signals
    -------
    sigSetDefaults(dict)
        Emits dictionary of current preview. Used to set the workflow pipeline according to the emitted dict

    Parameters
    ----------
    dim : int
        Dimensions of arrays in preview array deque. This is no longer used because array deque can hold arrays of
        different size.
    maxpreviews : int
        Maximum number of preview arrrays that can be held
    args
    kwargs
    """

    sigSetDefaults = QtCore.Signal(dict)

    def __init__(self, dim, maxpreviews=None, *args, **kwargs):
        super(PreviewViewer, self).__init__(*args, **kwargs)
        self.maxpreviews = maxpreviews if maxpreviews is not None else 40

        self.dim = dim

        self.previews = ArrayDeque(arrayshape=(dim, dim), maxlen=self.maxpreviews)
        self.datatrees = deque(maxlen=self.maxpreviews)
        self.data = deque(maxlen=self.maxpreviews)
        self.slice_numbers = deque(maxlen=self.maxpreviews)

        self.setOrientation(QtCore.Qt.Horizontal)
        self.functionform = QtGui.QStackedWidget()

        self.deleteButton = QtGui.QToolButton(self)
        self.deleteButton.setToolTip('Delete this preview')
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_36.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.deleteButton.setIcon(icon)

        self.setPipelineButton = QtGui.QToolButton(self)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_45.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.setPipelineButton.setIcon(icon)
        self.setPipelineButton.setToolTip('Set as pipeline')

        ly = QtGui.QVBoxLayout()
        ly.setContentsMargins(0, 0, 0, 0)
        ly.setSpacing(0)
        ly.addWidget(self.functionform)
        h = QtGui.QHBoxLayout()
        h.setContentsMargins(0, 0, 0, 0)
        h.addWidget(self.setPipelineButton)
        h.addWidget(self.deleteButton)
        ly.addLayout(h)
        panel = QtGui.QWidget(self)
        panel.resize(375,480)
        panel.setLayout(ly)
        self.setPipelineButton.hide()
        self.deleteButton.hide()

        self.imageview = ImageView(self)
        self.imageview.ui.roiBtn.setParent(None)
        self.imageview.ui.menuBtn.setParent(None)

        self.view_label = QtGui.QLabel(self)
        self.view_label.setText('No: ')
        self.view_number = QtGui.QSpinBox(self)
        self.view_number.setReadOnly(True)
        self.setButton = histDialogButton('Set', parent=self)
        self.setButton.connectToHistWidget(self.imageview.getHistogramWidget())

        self.view_number.setMaximum(5000) # Large enough number
        self.imageview.ui.gridLayout.addWidget(self.setButton, 1, 1, 1, 2)
        self.imageview.ui.gridLayout.addWidget(self.view_label, 2, 1, 1, 1)
        self.imageview.ui.gridLayout.addWidget(self.view_number, 2, 2, 1, 1)

        self.setCurrentIndex = self.imageview.setCurrentIndex
        self.addWidget(panel)
        self.addWidget(self.imageview)

        self.imageview.sigDeletePressed.connect(self.removePreview)
        self.setPipelineButton.clicked.connect(self.defaultsButtonClicked)
        self.deleteButton.clicked.connect(self.removePreview)
        self.imageview.sigTimeChanged.connect(self.indexChanged)


    @ QtCore.Slot(object, object)
    def indexChanged(self, index, time):
        """Slot connected to the ImageViews sigChanged"""
        try:
            self.functionform.setCurrentWidget(self.datatrees[index])
            self.view_number.setValue(self.slice_numbers[index])
        except IndexError as e:
            print('index {} does not exist'.format(index))

    # Could be leaking memory if I don't explicitly delete the datatrees that are being removed
    # from the previewdata deque but are still in the functionform widget? Hopefully python gc is taking good care of me
    def addPreview(self, image, funcdata, slice_number):
        """
        Adds a preview

        Parameters
        ----------
        image : ndarray
            Reconstructed image
        funcdata : dict
            Dictionary summarizing pipeline used for reconstruction
        slice_number : int
            Index of sinogram reconstructed
        """

        self.deleteButton.show()
        self.setPipelineButton.show()
        self.previews.appendleft(np.flipud(image))
        functree = DataTreeWidget()
        functree.setHeaderHidden(False)
        functree.setHeaderLabels(['Function', 'Params'])
        functree.setData(funcdata, hideRoot=True)
        functree.setSelectionMode(QtGui.QAbstractItemView.SingleSelection)
        functree.setSelectionBehavior(QtGui.QAbstractItemView.SelectItems)
        functree.setColumnWidth(0, 140)
        functree.setColumnWidth(1, 235)

        self.data.appendleft(funcdata)
        self.datatrees.appendleft(functree)
        self.slice_numbers.appendleft(slice_number)
        self.view_number.setValue(slice_number)
        self.functionform.addWidget(functree)
        levels = False if len(self.data) > 1 else True
        self.imageview.setImage(self.previews, autoRange=False, autoLevels=levels, autoHistogramRange=False)

        # disable autoscaling for histogram
        hist = self.imageview.getHistogramWidget()
        hist.vb.enableAutoRange(hist.vb.YAxis, False)

        self.functionform.setCurrentWidget(functree)

    def removePreview(self):
        """
        Removes the current preview
        """
        if len(self.previews) > 0:
            idx = self.imageview.currentIndex
            self.functionform.removeWidget(self.datatrees[idx])
            del self.previews[idx]
            del self.datatrees[idx]
            del self.data[idx]
            del self.slice_numbers[idx]
            if len(self.previews) == 0:
                self.imageview.clear()
                self.deleteButton.hide()
                self.setPipelineButton.hide()
            else:
                self.imageview.setImage(self.previews)

    def defaultsButtonClicked(self):
        """
        Emits the dict of current preview
        """
        current_data = self.data[self.imageview.currentIndex]
        self.sigSetDefaults.emit(current_data)


class Preview3DViewer(QtGui.QSplitter):
    """
    Viewer class to show 3D reconstruction previews, along with the function pipeline settings for the
    corresponding preview


    Attributes
    ----------
    volumviewer : widgets,volumeviewers.VolumeViewer
        VolumeViewer widget to render 3D preview reconstruction volume
    fdata : dict
        dict corresponding to the reconstruction functions
    pipelinetree : DataTree widget
        Datatree for displaying data dict
    data : ndarray
        Array of reconstructed volume

    Signals
    -------
    sigSetDefaults(dict)
        Emits dictionary of preview. Used to set the workflow pipeline according to the emitted dict

    """

    sigSetDefaults = QtCore.Signal(dict)

    def __init__(self, *args, **kwargs):
        super(Preview3DViewer, self).__init__(*args, **kwargs)
        self.setOrientation(QtCore.Qt.Horizontal)
        l = QtGui.QVBoxLayout()
        l.setContentsMargins(0, 0, 0, 0)
        self.pipelinetree = DataTreeWidget()
        self.pipelinetree.setHeaderHidden(False)
        self.pipelinetree.clear()

        self.setPipelineButton = QtGui.QToolButton(self)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_45.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.setPipelineButton.setIcon(icon)
        self.setPipelineButton.setToolTip('Set as pipeline')

        ly = QtGui.QVBoxLayout()
        ly.setContentsMargins(0, 0, 0, 0)
        ly.setSpacing(0)
        ly.addWidget(self.pipelinetree)
        h = QtGui.QHBoxLayout()
        h.setContentsMargins(0, 0, 0, 0)
        h.addWidget(self.setPipelineButton)
        ly.addLayout(h)
        panel = QtGui.QWidget(self)
        panel.resize(400, 480)
        panel.setLayout(ly)

        self.volumeviewer = VolumeViewer()

        self.addWidget(panel)
        self.addWidget(self.volumeviewer)

        self.data = None

        self.setPipelineButton.clicked.connect(lambda: self.sigSetDefaults.emit(self.data))
        self.setPipelineButton.hide()

        self.resize(800, 480)

        # self.pipelinetree.setColumnWidth(0, 140)
        # self.pipelinetree.setColumnWidth(1, 235)

    def setPreview(self, recon, funcdata):
        """
        Sets the 3D preview

        Parameters
        ----------
        recon : ndarray
            3D array of reconstructed volume
        funcdata : dict
            Dictionary summarizing pipeline used for reconstruction
        """

        self.pipelinetree.setData(funcdata, hideRoot=True)
        self.data = funcdata
        self.pipelinetree.setColumnWidth(0, 130)
        self.pipelinetree.setColumnWidth(1, 245)
        self.pipelinetree.show()
        self.volumeviewer.setVolume(vol=recon)
        self.setPipelineButton.show()


class RunConsole(QtGui.QTabWidget):
    """
    Class to output status of a running job, and cancel the job.  Has tab for local run settings
    and can add tabs tab for remote job settings.

    Attributes
    ----------
    local_console : QtGui.QWidget
        Widget for console used when running local reconstructions
    """

    icon = QtGui.QIcon()
    icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_51.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
    icon_clear = QtGui.QIcon()
    icon_clear.addPixmap(QtGui.QPixmap("xicam/gui/icons_57.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)

    def __init__(self, parent=None):
        super(RunConsole, self).__init__(parent=parent)
        self.setTabPosition(QtGui.QTabWidget.West)
        # Text Browser for local run console
        self.local_console, self.local_cancelButton, self.local_clearButton = self.addConsole('Local')
        self.local_clearButton.clicked.connect(self.local_console.clear)

    def addConsole(self, name):
        """
        Adds a new console, This will come in handy when running remote operations (ie adding a console for remote
        location). Will probably need to create an attribute (dict) for holding the names, consoles when adding consoles

        Parameters
        ----------
        name : str
            Name to be used in tab for console added
        """

        # TODO: finish adding message clear button
        console = QtGui.QTextEdit()
        button = QtGui.QToolButton()
        button_clear = QtGui.QToolButton()

        console.setObjectName(name)
        console.setReadOnly(True)
        console.setSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Preferred)
        button.setIcon(self.icon)
        button.setIconSize(QtCore.QSize(24, 24))
        button.setFixedSize(32, 32)
        button.setToolTip('Cancel running process')
        button_clear.setIcon(self.icon_clear)
        button_clear.setIconSize(QtCore.QSize(24,24))
        button_clear.setFixedSize(32,32)
        button_clear.setToolTip('Clear console log')

        w = QtGui.QWidget()
        w.setSizePolicy(QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Preferred)
        w.setContentsMargins(0, 0, 0, 0)
        l = QtGui.QGridLayout()
        l.setContentsMargins(0, 0, 0, 0)
        l.setSpacing(0)
        l.addWidget(console, 0, 0, 3, 2)
        l.addWidget(button, 0, 2, 1, 1)
        l.addWidget(button_clear,1,2,1,1)
        w.setLayout(l)
        self.addTab(w, console.objectName())
        return console, button, button_clear

    def log2local(self, msg):
        """
        Logs a message to the local console. If adding new consoles there will have to be a way to dynamically create
        a function like this for the added console.
        """
        text = self.local_console.toPlainText()
        try:
            if '\n' not in msg:
                self.local_console.setText(msg + '\n\n' + text)
            else:
                topline = text.splitlines()[0]
                tail = '\n'.join(text.splitlines()[1:])
                self.local_console.setText(topline + msg + tail)
        except Exception:
            pass


class ArrayDeque(deque):
    """
    Class for a numpy array deque where arrays can be appended on both ends.

    Parameters
    ----------
    arraylist : list of ndarrays, optional
        List of ndarrays to initialize
    arrayshape : tuple, optional
        Shape of ndarrays to be held. This is not lenient and not needed since arrays of different sizes can be added
    dtype : type
        Type of array data
    maxlen : int
        Maximum number of arrays that can be held in ArrayDeque
    """

    def __init__(self, arraylist=[], arrayshape=None, dtype=None, maxlen=None):
        # perhaps will need to add check of datatype everytime a new array is added with extend, append, etc??
        if not arraylist and not arrayshape:
            raise ValueError('One of arraylist or arrayshape must be specified')

        super(ArrayDeque, self).__init__(iterable=arraylist, maxlen=maxlen)

        self._shape = [len(self)]
        self._dtype = dtype

        if arraylist:
            # if False in [np.array_equal(arraylist[0].shape, array.shape) for array in arraylist[1:]]:
            #     raise ValueError('All arrays in arraylist must have the same dimensions')
            # elif False in [arraylist[0].dtype == array.dtype for array in arraylist[1:]]:
            #     raise ValueError('All arrays in arraylist must have the same data type')
            map(self._shape.append, arraylist[0].shape)
        elif arrayshape:
            map(self._shape.append, arrayshape)

        self.ndim = len(self._shape)

    @property
    def shape(self):
        """
        Return the shape of the deque based on number of arrays held
        """
        self._shape[0] = len(self)
        return self._shape

    @property
    def size(self):
        """
        Return the size of the array based on number of arrays held
        """
        return np.product(self.shape)

    @property
    def dtype(self):
        """
        Return the dataype of the array's in deque
        """
        if self._dtype is None and self.shape[0]:
            self._dtype = self.__getitem__(0).dtype
        return self._dtype

    @property
    def max(self):
        """
        Return the maximum value
        """
        return np.max(max(self, key=lambda x:np.max(x)))

    @property
    def min(self):
        """
        Return the minimum value
        """
        return np.min(min(self, key=lambda x:np.min(x)))

    def append(self, arr):
        """
        Appends an array to the end of the array deque
        """

        # if arr.shape != tuple(self.shape[1:]):
        #     raise ValueError('Array shape must be {0}, got shape {1}'.format(self.shape[1:], arr.shape))
        # if self.dtype is not None and arr.dtype != self.dtype:
        #     raise ValueError('Array must be of type {}'.format(self.dtype))
        super(ArrayDeque, self).append(arr)

    def appendleft(self, arr):
        """
        Appends an array to the beginning of the array deque
        """

        # if arr.shape != tuple(self.shape[1:]):
        #     raise ValueError('Array shape must be {0}, got shape {1}'.format(self.shape[1:], arr.shape))
        # if self.dtype is not None and arr.dtype != self.dtype:
        #     raise ValueError('Array must be of type {}'.format(self.dtype))
        super(ArrayDeque, self).appendleft(arr)

    def __getitem__(self, item):
        """
        Override slicing
        """
        if type(item) is list and isinstance(item[0], slice):
            dq_item = item.pop(0)
            if isinstance(dq_item, slice):
                dq_item = dq_item.stop if dq_item.stop is not None else dq_item.start if dq_item.start is not None else 0
            return super(ArrayDeque, self).__getitem__(dq_item).__getitem__(item)
        else:
            return super(ArrayDeque, self).__getitem__(item)

    def transpose(self, ax):
        return self

# Testing
if __name__ == '__main__':
    import sys

    app = QtGui.QApplication(sys.argv)
    w = RunConsole()
    def foobar():
        for i in range(10000):
            w.log2local('Line {}\n\n'.format(i))
            # time.sleep(.1)
    w.local_cancelButton.clicked.connect(foobar)
    w.setWindowTitle("Test this thing")
    w.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2015, Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
# -----------------------------------------------------------------------------
# vispy: gallery 2
#
# Adapted for use as a widget by Ron Pandolfi
# volumeViewer.getHistogram method borrowed from PyQtGraph

"""
Example volume rendering

Controls:

* 1  - toggle camera between first person (fly), regular 3D (turntable) and
       arcball
* 2  - toggle between volume rendering methods
* 3  - toggle between stent-CT / brain-MRI image
* 4  - toggle between colormaps
* 0  - reset cameras
* [] - decrease/increase isosurface threshold

With fly camera:

* WASD or arrow keys - move around
* SPACE - brake
* FC - move up-down
* IJKL or mouse - look around
"""

from PySide import QtGui
from vispy import scene

from pipeline import loader, msg
# TODO refactor general widgets to be part of plugins.widgets to be shared in a more organized fashion
from xicam.widgets.volumeviewers import VolumeVisual, VolumeViewer
from xicam.widgets.imageviewers import StackViewer


class ThreeDViewer(QtGui.QWidget, ):
    def __init__(self, paths, parent=None):
        super(ThreeDViewer, self).__init__(parent=parent)

        self.combo_box = QtGui.QComboBox(self)
        self.combo_box.addItems(['Image Stack', '3D Volume'])
        self.stack_viewer = StackViewer()
        self.volume_viewer = VolumeViewer()

        self.view_stack = QtGui.QStackedWidget(self)
        self.view_stack.addWidget(self.stack_viewer)
        self.view_stack.addWidget(self.volume_viewer)

        hlayout = QtGui.QHBoxLayout()
        self.subsample_spinbox = QtGui.QSpinBox()
        self.subsample_label = QtGui.QLabel('Subsample Level:')
        self.loadVolumeButton = QtGui.QToolButton()
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_45.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.loadVolumeButton.setIcon(icon)
        self.loadVolumeButton.setToolTip('Generate Volume')
        hlayout.addWidget(self.combo_box)
        hlayout.addSpacing(2)
        hlayout.addWidget(self.subsample_label)
        hlayout.addWidget(self.subsample_spinbox)
        hlayout.addWidget(self.loadVolumeButton)
        hlayout.addStretch()
        layout = QtGui.QVBoxLayout(self)
        layout.addLayout(hlayout)
        layout.addWidget(self.view_stack)

        self.subsample_spinbox.hide()
        self.subsample_spinbox.setValue(8)
        self.subsample_label.hide()
        self.loadVolumeButton.hide()

        self.stack_image = loader.StackImage(paths)
        self.volume = None
        self.stack_viewer.setData(self.stack_image)
        self.combo_box.activated.connect(self.view_stack.setCurrentIndex)
        self.view_stack.currentChanged.connect(self.toggleInputs)
        self.loadVolumeButton.clicked.connect(self.loadVolume)

    def toggleInputs(self, index):
        if self.view_stack.currentWidget() is self.volume_viewer:
            self.subsample_label.show()
            self.subsample_spinbox.show()
            self.loadVolumeButton.show()
        else:
            self.subsample_label.hide()
            self.subsample_spinbox.hide()
            self.loadVolumeButton.hide()

    def loadVolume(self):
        msg.showMessage('Generating volume...', timeout=5)
        level = self.subsample_spinbox.value()
        self.volume = self.stack_image.asVolume(level=level)
        self.volume_viewer.setVolume(vol=self.volume, slicevol=False)
        msg.clearMessage()



import logging
import inspect
import sys
import time

statusbar = None # Must be registered to output to a ui status bar
progressbar = None

stdch = logging.StreamHandler(sys.stdout)

guilogcallable = None

DEBUG = logging.DEBUG           # 10
INFO = logging.INFO             # 20
WARNING = logging.WARNING       # 30
ERROR = logging.ERROR           # 40
CRITICAL = logging.CRITICAL     # 50

logbacklog=[]


def showProgress(value, min=0, max=100):
    if progressbar:
        progressbar.show()
        progressbar.setRange(min, max)
        progressbar.setValue(value)


def showBusy():
    if progressbar:
        progressbar.show()
        progressbar.setRange(0, 0)


def hideBusy():
    if progressbar:
        progressbar.hide()
        progressbar.setRange(0, 100)


hideProgress = hideBusy


def showMessage(s,timeout=0):
    if statusbar is not None:
        statusbar.showMessage(s, timeout * 1000)

    logMessage(s)

def logMessage(stuple,level=INFO,loggername=None,timestamp=None,image=None,suppressreprint=False):

    # ATTENTION: loggername is 'intelligently' determined with inspect. You probably want to leave it None.
    if loggername is not None:
        loggername = inspect.stack()[1][3]
    logger = logging.getLogger(loggername)
    try:
        stdch.setLevel(level)
    except ValueError:
        print stuple,level
    logger.addHandler(stdch)

    if timestamp is None: timestamp = time.asctime()

    if type(stuple) is not tuple:
        stuple=[stuple]

    stuple = (unicode(s) for s in stuple)

    s = ' '.join(stuple)
    m = timestamp +'\t'+unicode(s)

    logger.log(level,m)
    if guilogcallable:
        guilogcallable(level,timestamp,s,image)
    else:
        global logbacklog
        logbacklog.append({'stuple':s,'level':level,'loggername':loggername,'timestamp':timestamp,'image':image})
    try:
        if not suppressreprint: print m
    except UnicodeEncodeError:
        print 'A unicode string could not be written to console. Some logging will not be displayed.'

def flushbacklog():
    global logbacklog
    for l in logbacklog:
        l['suppressreprint']=True
        logMessage(**l)
    logbacklog=[]

def clearMessage():
    statusbar.clearMessage()




# --coding: utf-8 --

from PySide import QtGui, QtCore
from PySide.QtCore import Qt
import numpy as np
import pyqtgraph as pg
from pipeline import loader, cosmics, integration, peakfinding, center_approx, variationoperators, pathtools, writer
from xicam import config, ROI, debugtools, toolbar
from fabio import edfimage
import os
import pyqtgraph.parametertree.parameterTypes as pTypes
from pyqtgraph.parametertree import Parameter, ParameterTree, ParameterItem, registerParameterType
from xicam import dialogs
from xicam import xglobals
import scipy
from pipeline import variation
from xicam import threads
from pipeline import msg
from scipy.ndimage import morphology
from pyFAI import calibrant

class OOMTabItem(QtGui.QWidget):
    sigLoaded = QtCore.Signal()

    def __init__(self, itemclass=None, *args, **kwargs):
        """
        A collection of references that can be used to reconstruct an object dynamically and dispose of it when unneeded
        :type paths: list[str]
        :type experiment: config.experiment
        :type parent: main.MyMainWindow
        :type operation:
        :return:
        """
        super(OOMTabItem, self).__init__()

        self.itemclass = itemclass
        self.args = args
        self.kwargs = kwargs
        self.saved = None

        self.isloaded = False
        self.setLayout(QtGui.QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

    def load(self):
        """
        load this tab; rebuild the viewer
        """
        if not self.isloaded:
            if 'operation' in self.kwargs:
                if self.kwargs['operation'] is not None:
                    msg.logMessage(self.kwargs['src'], msg.DEBUG)
                    imgdata = [loader.loadimage(path) for path in self.kwargs['src']]
                    imgdata = self.kwargs['operation'](imgdata)
                    dimg = loader.datadiffimage2(data=imgdata)
                    self.kwargs['dimg'] = dimg

            self.widget = self.itemclass(*self.args, **self.kwargs)

            if hasattr(self.widget,'restore'):
                self.widget.restore(*self.saved)

            self.layout().addWidget(self.widget)

            self.isloaded = True

            self.sigLoaded.emit()
            msg.logMessage(('Loaded:', self.itemclass),msg.DEBUG)

    def unload(self):
        """
        orphan the tab widget and queue them for deletion. Mwahahaha.
        Try to save values return from widget's save; restore values with widget's 'restore' method
        """

        if self.isloaded:
            if hasattr(self.widget,'save'):
                self.saved = self.widget.save()

            self.widget.deleteLater()
            self.widget = None
            self.isloaded = False


class dimgViewer(QtGui.QWidget):

    def __init__(self, dimg=None, src=None, plotwidget=None, toolbar=None, **kwargs):
        """
        A tab containing an imageview. Also manages functionality connected to a specific tab (masking/integration)
        :param imgdata:
        :param experiment:
        :return:
        """
        super(dimgViewer, self).__init__()

        self.region = None
        self.maskROI = None
        self.layout = QtGui.QStackedLayout(self)
        self.plotwidget = plotwidget
        self.toolbar = toolbar

        if dimg is not None:
            self.dimg = dimg
        else:
            try:
                self.dimg = loader.loaddiffimage(src)
            except RuntimeError: # unknown max recursion error
                self.dimg = loader.loaddiffimage(src)

        # Make an imageview for the image
        self.imgview = ImageView(self,actionLog_Intensity=self.toolbar.actionLog_Intensity)
        self.imageitem = self.imgview.getImageItem()
        self.graphicslayoutwidget = self.imgview
        self.imgview.ui.roiBtn.setParent(None)

        self.viewbox = self.imageitem.getViewBox()
        # self.imgview.view.removeItem(self.imgview.roi)
        # self.imgview.roi.parent = None
        # self.imgview.roi.deleteLater()
        # self.imgview.view.removeItem(self.imgview.normRoi)       #Should uncomment?
        # self.imgview.normRoi.parent = None
        # self.imgview.normRoi.deleteLater()
        self.viewbox.invertY(False)
        reset = QtGui.QAction('Reset', self.imgview.getHistogramWidget().item.vb.menu)
        self.imgview.getHistogramWidget().item.vb.menu.addAction(reset)
        reset.triggered.connect(self.resetLUT)

        # self.imgview.getHistogramWidget().plot.setLogMode(True,False)


        # self.threads = dict()

        # cross hair
        linepen = pg.mkPen('#FFA500')
        self.vLine = pg.InfiniteLine(angle=90, movable=False, pen=linepen)
        self.hLine = pg.InfiniteLine(angle=0, movable=False, pen=linepen)
        self.vLine.setVisible(False)
        self.hLine.setVisible(False)
        self.viewbox.addItem(self.vLine, ignoreBounds=True)
        self.viewbox.addItem(self.hLine, ignoreBounds=True)

        # Add a thin border to the image so it is visible on black background
        self.imageitem.border = pg.mkPen('w')

        self.coordslabel = QtGui.QLabel(' ')
        self.coordslabel.setMinimumHeight(16)
        self.imgview.layout().addWidget(self.coordslabel)
        self.imgview.setStyleSheet("background-color: rgba(0,0,0,0%)")
        self.coordslabel.setAlignment(Qt.AlignHCenter | Qt.AlignBottom)
        self.coordslabel.setStyleSheet("background-color: rgba(0,0,0,0%)")
        self.graphicslayoutwidget.scene.sigMouseMoved.connect(self.mouseMoved)
        self.layout.setStackingMode(QtGui.QStackedLayout.StackAll)
        self.coordslabel.mouseMoveEvent = self.graphicslayoutwidget.mouseMoveEvent
        self.coordslabel.mousePressEvent = self.graphicslayoutwidget.mousePressEvent
        self.coordslabel.mouseReleaseEvent = self.graphicslayoutwidget.mouseReleaseEvent
        self.coordslabel.mouseDoubleClickEvent = self.graphicslayoutwidget.mouseDoubleClickEvent
        self.coordslabel.mouseGrabber = self.graphicslayoutwidget.mouseGrabber
        self.coordslabel.wheelEvent = self.graphicslayoutwidget.wheelEvent
        self.coordslabel.leaveEvent = self.graphicslayoutwidget.leaveEvent
        self.coordslabel.enterEvent = self.graphicslayoutwidget.enterEvent
        self.coordslabel.setMouseTracking(True)

        menu = self.viewbox.menu
        setcenter = QtGui.QAction('Set Center',menu)
        setcenter.setObjectName('setcenter')
        setcenter.triggered.connect(self.setcenter)
        menu.addAction(setcenter)


        self.centerplot = pg.ScatterPlotItem()
        self.viewbox.addItem(self.centerplot)

        self.sgToverlay = pg.ScatterPlotItem()
        self.viewbox.addItem(self.sgToverlay)

        self.sgRoverlay = pg.ScatterPlotItem()
        self.viewbox.addItem(self.sgRoverlay)

        # Make a layout for the tab
        backwidget = QtGui.QWidget()
        self.layout.addWidget(backwidget)
        self.backlayout = QtGui.QHBoxLayout(backwidget)
        self.backlayout.setContentsMargins(0, 0, 0, 0)
        self.backlayout.addWidget(self.graphicslayoutwidget)

        # Add the Log Intensity check button to the context menu and wire up
        # self.imgview.buildMenu()
        # menu = self.viewbox.menu
        # self.actionLogIntensity = QAction('Log Intensity', menu, checkable=True)
        # self.actionLogIntensity.triggered.connect(self.logintensity)
        # menu.addAction(self.actionLogIntensity)
        # self.imgview.buildMenu()

        # Add a placeholder image item for the mask to the viewbox
        self.maskimage = pg.ImageItem(opacity=.25)
        self.viewbox.addItem(self.maskimage)

        # Add a placeholder image item for the calibration to the viewbox
        self.calibrantoverlay = pg.ImageItem(opacity=.25)
        self.viewbox.addItem(self.calibrantoverlay)

        try:
            energy = self.dimg.headers['Beamline Energy']
            if energy is not None:
                config.activeExperiment.setvalue('Energy', energy)
        except (AttributeError, TypeError, KeyError):
            msg.logMessage('Warning: Energy could not be determined from headers',msg.WARNING)

        # Force cache the detector
        _ = self.dimg.detector

        # Cache radial integration
        if self.dimg is not None:
            if self.dimg.rawdata is not None:
                self.redrawimage()

                # self.q, self.radialprofile = self.dimg.radialintegrate

                # Force cache the detector
                # _ = self.dimg.detector
                if not self.loadLUT():
                    self.resetLUT()

                if config.activeExperiment.iscalibrated:
                    self.replot()
                    self.drawcenter()

        self.imgview.getHistogramWidget().item.sigLevelChangeFinished.connect(self.cacheLUT)
        self.imgview.getHistogramWidget().item.gradient.sigGradientChangeFinished.connect(self.cacheLUT)


    def mousePressEvent(self,ev):
        super(dimgViewer, self).mousePressEvent(ev)
        self.lastclick = self.imageitem.mapFromScene(ev.pos())

    def setcenter(self,*args,**kwargs):
        config.activeExperiment.center=self.lastclick.x(),self.lastclick.y()
        self.redrawimage()
        self.replot()

    def loadLUT(self):

        if xglobals.LUT is not None:
            hist = self.imgview.getHistogramWidget().item
            hist.setLevels(*xglobals.LUTlevels)
            hist.gradient.restoreState(xglobals.LUTstate)
            return True
        return False

    def resetLUT(self):

        msg.logMessage(('Levels:', self.imgview.getHistogramWidget().item.getLevels()),msg.DEBUG)
        # if self.imgview.getHistogramWidget().item.getLevels()==(0,1.):
        Lmax = np.nanmax(self.dimg.rawdata)

        self.imgview.autoLevels()
        msg.logMessage(('Levels set:', self.imgview.getHistogramWidget().item.getLevels()),msg.DEBUG)

    def cacheLUT(self):
        hist = self.imgview.getHistogramWidget().item
        xglobals.LUTlevels = hist.getLevels()
        xglobals.LUTstate = hist.gradient.saveState()
        xglobals.LUT = hist.getLookupTable(img=self.imageitem.image)

    # def send1Dintegration(self):
    # self.cache1Dintegration.emit(self.q, self.radialprofile)

    def removeROI(self, evt):

        # evt.scene().removeItem(evt)
        self.viewbox.removeItem(evt)
        self.replot()
        # evt.deleteLater()

        # self.viewbox.scene().removeItem(evt)

    # @property
    # def isChiPlot(self):
    #     return self.plotwidget.currentIndex()

    def mouseMoved(self, evt):
        """
        when the mouse is moved in the viewer, translate the crosshair, recalculate coordinates
        """
        pos = evt  ## using signal proxy turns original arguments into a tuple
        if self.viewbox.sceneBoundingRect().contains(pos):
            mousePoint = self.viewbox.mapSceneToView(pos)
            self.vLine.setPos(mousePoint.x())
            self.hLine.setPos(mousePoint.y())
            if (0 < mousePoint.x() < self.imageitem.image.shape[0]) & (
                            0 < mousePoint.y() < self.imageitem.image.shape[1]):  # within bounds
                # angstrom=QChar(0x00B5)
                if config.activeExperiment.iscalibrated:
                    x = mousePoint.x()
                    y = mousePoint.y()

                    data = self.dimg.transformdata

                    # if iscake:
                    # q = pixel2cake(x, y, self.dimg)
                    #
                    # elif isremesh:
                    # return
                    # else:
                    #     q = pixel2q(x, y, self.dimg.experiment)
                    # print x,y,self.dimg.data[int(x),int(y)],self.getq(x,y),self.getq(None,y),self.getq(x,None,),np.sqrt((x - self.dimg.experiment.center[0]) ** 2 + (y - self.dimg.experiment.center[1]) ** 2)
                    self.coordslabel.setText(u"<div style='font-size: 12pt;background-color:#111111;'>x=%0.1f,"
                                             u"   <span style=''>y=%0.1f</span>,   <span style=''>I=%0.0f</span>,"
                                             u"  q=%0.3f \u212B\u207B\u00B9,  q<sub>z</sub>=%0.3f \u212B\u207B\u00B9,"
                                             u"  q<sub>\u2225\u2225</sub>=%0.3f \u212B\u207B\u00B9,"
                                             u"  d=%0.3f nm,"
                                             u"  \u03B8=%.2f</div>" % (
                                                 x,
                                                 y,
                                                 data[int(x),
                                                      int(y)],
                                                 self.getq(x, y),
                                                 self.getq(x, y, 'z'),
                                                 self.getq(x, y, 'parallel'),
                                                 2 * np.pi / self.getq(x, y) / 10,
                                                 360. / (2 * np.pi) * np.arctan2(self.getq(x, y, 'z'),
                                                                                 self.getq(x, y, 'parallel'))))
                    # np.sqrt((x - self.dimg.experiment.center[0]) ** 2 + (
                    # y - self.dimg.experiment.center[1]) ** 2)))
                    # ,  r=%0.1f
                    if self.plotwidget is not None: #for timeline
                        self.plotwidget.movPosLine(self.getq(x, y),
                                                   self.getq(x, y, mode='parallel'),
                                                   self.getq(x, y, mode='z'))

                else:
                    self.coordslabel.setText(u"<div style='font-size: 12pt;background-color:#111111;'>x=%0.1f,"
                                             u"   <span style=''>y=%0.1f</span>,   <span style=''>I=%0.0f</span>,"
                                             u"  Calibration Required...</div>" % (
                                                 mousePoint.x(),
                                                 mousePoint.y(),
                                                 self.dimg.data[int(mousePoint.x()),
                                                                int(mousePoint.y())],
                                             ))

                    # self.coordslabel.setVisible(True)

            else:
                self.coordslabel.setText(u"<div style='font-size:12pt;background-color:#111111;'>&nbsp;</div>")
                if hasattr(self.plotwidget, 'qintegration'):
                    self.plotwidget.qintegration.posLine.hide()

    def getq(self, x, y, mode=None):  # This is a mess...rewrite it sometime
        iscake = self.toolbar.actionCake.isChecked()
        isremesh = self.toolbar.actionRemeshing.isChecked()

        if iscake:
            cakeq = self.dimg.cakeqx
            cakechi = self.dimg.cakeqy
            if mode is not None:
                if mode == 'parallel':
                    return cakeq[int(y)] * np.sin(np.radians(cakechi[int(x)])) / 10.
                elif mode == 'z':
                    return cakeq[int(y)] * np.cos(np.radians(cakechi[int(x)])) / 10.
            else:
                return cakeq[int(y)] / 10.

        elif isremesh:
            remeshqpar = self.dimg.remeshqx
            remeshqz = self.dimg.remeshqy
            if mode is not None:
                if mode == 'parallel':
                    return remeshqpar[int(x), int(y)] / 10.
                elif mode == 'z':
                    return -remeshqz[int(x), int(y)] / 10.
            else:
                return np.sqrt(remeshqz[int(x), int(y)] ** 2 + remeshqpar[int(x), int(y)] ** 2) / 10.

        else:
            center = config.activeExperiment.center
            if mode is not None:
                if mode == 'z':
                    x = center[0]
                if mode == 'parallel':
                    y = center[1]

            r = np.sqrt((x - center[0]) ** 2 + (y - center[1]) ** 2)
            theta = np.arctan2(r * config.activeExperiment.getvalue('Pixel Size X'),
                               config.activeExperiment.getvalue('Detector Distance'))
            wavelength = config.activeExperiment.getvalue('Wavelength')
            q = 4 * np.pi / wavelength * np.sin(theta / 2) * 1e-10
            if mode == 'parallel' and x < center[0]:
                return -q
            if mode == 'z' and y < center[1]:
                return -q
            return q

    def leaveEvent(self, evt):
        """
        hide crosshair and coordinates when mouse leaves viewer
        """
        self.hLine.setVisible(False)
        self.vLine.setVisible(False)
        # self.coordslabel.setVisible(False)
        if self.plotwidget:
            self.plotwidget.hidePosLine()

    def enterEvent(self, evt):
        """
        show crosshair and coordinates when mouse enters viewer
        """
        self.hLine.setVisible(True)
        self.vLine.setVisible(True)
        if hasattr(self.plotwidget, 'qintegration'):
            self.plotwidget.qintegration.posLine.setVisible(True)

    def redrawimage(self, returnimg=False):
        """
        redraws the diffraction image, checking drawing modes (log, symmetry, mask, cake)
        """

        self.dimg.cakemode = self.iscake
        self.dimg.remeshmode = self.isremesh
        self.dimg.radialsymmetrymode = self.isradialsymmetry
        self.dimg.mirrorsymmetrymode = self.ismirrorsymmetry
        self.dimg.logscale = self.islogintensity

        if returnimg:
            return self.dimg
        else:
            self.imgview.setImage(self.dimg)

        self.drawcenter()

        self.replot()

        if self.ismaskshown:
            self.maskoverlay()
        else:
            self.maskimage.clear()

        # if not iscake and not isremesh:
        #     self.imageitem.setRect(QtCore.QRect(0, 0, self.dimg.rawdata.shape[0], self.dimg.rawdata.shape[1]))

    def getcenter(self):
        if self.isremesh:
            remeshqpar = self.dimg.remeshqy
            remeshqz = self.dimg.remeshqx
            q = remeshqpar ** 2 + remeshqz ** 2
            center = np.where(q == q.min())
            return zip(*center)[0]
        else:
            return config.activeExperiment.center

    @property
    def islogintensity(self):
        return self.toolbar.actionLog_Intensity.isChecked()

    @property
    def isradialsymmetry(self):
        return self.toolbar.actionRadial_Symmetry.isChecked()

    @property
    def ismirrorsymmetry(self):
        return self.toolbar.actionMirror_Symmetry.isChecked()

    @property
    def ismaskshown(self):
        return self.toolbar.actionShow_Mask.isChecked()

    @property
    def iscake(self):
        return self.toolbar.actionCake.isChecked()

    @property
    def isremesh(self):
        return self.toolbar.actionRemeshing.isChecked()

    def arccut(self):

        arc = ROI.ArcROI(self.getcenter(), 500)
        arc.sigRegionChangeFinished.connect(self.replot)
        self.viewbox.addItem(arc)
        self.replot()
        arc.sigRemoveRequested.connect(self.removeROI)
        xglobals.lastroi = (arc, self.imageitem)

    def linecut(self):
        """
        toggles the line cut
        """
        # self.viewbox.removeItem(self.region)
        # self.parentwindow.difftoolbar.actionVertical_Cut.setChecked(False)
        # self.parentwindow.difftoolbar.actionHorizontal_Cut.setChecked(False)
        # if self.parentwindow.difftoolbar.actionLine_Cut.isChecked():
        region = ROI.LineROI(
            self.getcenter(),
            [self.getcenter()[0], -self.dimg.transformdata.shape[0]], 5, removable=True)
        region.sigRemoveRequested.connect(self.removeROI)
        self.viewbox.addItem(region)
        self.replot()
        region.sigRegionChangeFinished.connect(self.replot)
        xglobals.lastroi = (region, self.imageitem)
        # else:
        # #self.viewbox.removeItem(self.region)
        # self.region = None
        # self.replot()

    def verticalcut(self):
        # self.viewbox.removeItem(self.region)
        # self.parentwindow.difftoolbar.actionLine_Cut.setChecked(False)
        # self.parentwindow.difftoolbar.actionHorizontal_Cut.setChecked(False)
        # if self.parentwindow.difftoolbar.actionVertical_Cut.isChecked():
        # try:
        # self.viewbox.removeItem(self.region)
        # except AttributeError:
        #         print('Attribute error in verticalcut')
        region = ROI.LinearRegionItem(orientation=pg.LinearRegionItem.Vertical, brush=pg.mkBrush('#00FFFF32'),
                                      bounds=[0, self.dimg.transformdata.shape[0]],
                                      values=[self.getcenter()[0] - 10,
                                              10 + self.getcenter()[0]])
        for line in region.lines:
            line.setPen(pg.mkPen('#00FFFF'))
        region.sigRegionChangeFinished.connect(self.replot)
        region.sigRemoveRequested.connect(self.removeROI)
        self.viewbox.addItem(region)
        self.replot()
        xglobals.lastroi = (region, self.imageitem)
        # else:
        # #self.viewbox.removeItem(self.region)
        #     self.region = None
        # self.replot()

    def horizontalcut(self):
        # self.parentwindow.difftoolbar.actionVertical_Cut.setChecked(False)
        # self.parentwindow.difftoolbar.actionLine_Cut.setChecked(False)
        # self.viewbox.removeItem(self.region)
        # if self.parentwindow.difftoolbar.actionHorizontal_Cut.isChecked():
        # try:
        # self.viewbox.removeItem(self.region)
        # except AttributeError:
        #         print('Attribute error in horizontalcut')
        region = ROI.LinearRegionItem(orientation=pg.LinearRegionItem.Horizontal, brush=pg.mkBrush('#00FFFF32'),
                                      bounds=[0, self.dimg.transformdata.shape[1]],
                                      values=[self.getcenter()[1]-10,
                                              10 + self.getcenter()[1]])
        for line in region.lines:
            line.setPen(pg.mkPen('#00FFFF'))
        region.sigRegionChangeFinished.connect(self.replot)
        region.sigRemoveRequested.connect(self.removeROI)
        self.viewbox.addItem(region)

        self.replot()
        xglobals.lastroi = (region, self.imageitem)
        # else:
        # #self.viewbox.removeItem(self.region)
        #     self.region = None
        # self.replot()

    def removecosmics(self):
        c = cosmics.cosmicsimage(self.dimg.rawdata)
        c.run(maxiter=4)
        config.activeExperiment.addtomask(1-c.mask)
        # self.maskoverlay()

    @debugtools.timeit
    def findcenter(self, skipdraw=False):
        # Auto find the beam center
        self.dimg.findcenter()
        if not skipdraw:
            self.drawcenter()
            self.replot()

    def drawcenter(self):

        center = self.getcenter()
        # Mark the center
        if self.centerplot is not None: self.centerplot.clear()
        self.centerplot.setData([center[0]], [center[1]], pen=None, symbol='o',
                                brush=pg.mkBrush('#FFA500'))
        # self.viewbox.addItem(self.centerplot)

        # Move Arc ROIs to center


        for item in self.viewbox.addedItems:
            # print item
            if issubclass(type(item), ROI.ArcROI):
                item.setPos(center)

    def simulatecalibrant(self, calibrantkey):
        ai = config.activeExperiment.getAI()
        c = calibrant.ALL_CALIBRANTS[calibrantkey]
        c.set_wavelength(ai.wavelength)
        fakecalibrationimg = c.fake_calibration_image(ai, shape=self.dimg.displaydata.shape[::-1], Imax=255, U=0, V=0, W=0.00001).T

        self.calibrantoverlay.setImage(np.dstack((np.ones_like(fakecalibrationimg), fakecalibrationimg, np.zeros_like(fakecalibrationimg), fakecalibrationimg)).astype(np.float),
            opacity=.5)
        self.maskimage.setLevels([0, 1])

    @debugtools.timeit
    def calibrate(self, algorithm, calibrant):

        algorithm(self.dimg, calibrant)

        self.replot()
        self.drawcenter()

    @debugtools.timeit
    def refinecenter(self):
        # Force cache the detector
        # _=self.dimg.detector

        cen = center_approx.refinecenter(self.dimg)
        config.activeExperiment.center = cen
        self.drawcenter()

    def getROIs(self):
        return [roi for roi in self.viewbox.addedItems if hasattr(roi, 'isdeleting') and not roi.isdeleting]

    def replot(self):
        self.plotwidget.replot()
        #widget(self.plotwidget.currentIndex()).replot(self.dimg,self.getROIs(),self.imageitem)
        # if self.plotwidget.currentIndex() == 0:
        #     self.plotwidget.qintegration.clear()
        #
        #     # self.replotprimary()
        #     self.plotwidget.qintegration.qLine = pg.InfiniteLine(angle=90, movable=False, pen=pg.mkPen('#FFA500'))
        #     self.plotwidget.qintegration.qLine.setVisible(False)
        #     self.plotwidget.qintegration.addItem(self.plotwidget.qintegration.qLine)
        #     self.replotq()
        # elif self.plotwidget.currentIndex() == 1:
        #     self.plotwidget.chiintegration.clear()
        #     self.plotwidget.chiintegration.chiLine = pg.InfiniteLine(angle=90, movable=False, pen=pg.mkPen('#FFA500'))
        #     self.plotwidget.chiintegration.chiLine.setVisible(False)
        #     self.plotwidget.chiintegration.addItem(self.plotwidget.chiintegration.chiLine)
        #     self.replotchi()

    # def replotchi(self):
    #     if not config.activeExperiment.iscalibrated:
    #         return None
    #
    #     cut = None
    #
    #
    #
    #     data = self.dimg.transformdata
    #
    #     ai = config.activeExperiment.getAI().getPyFAI()
    #     xglobals.pool.apply_async(integration.chiintegratepyFAI,
    #                               args=(self.dimg.data, self.dimg.mask, ai, self.iscake),
    #                               callback=self.chiintegrationrelay)
    #     # pipeline.integration.chiintegratepyFAI(self.dimg.data, self.dimg.mask, ai, precaked=self.iscake)
    #
    #     for roi in self.viewbox.addedItems:
    #         # try:
    #             if hasattr(roi, 'isdeleting'):
    #                 if not roi.isdeleting:
    #                     print type(roi)
    #                     cut = None
    #                     if issubclass(type(roi), pg.ROI) or issubclass(type(roi), pg.LinearRegionItem):
    #                         cut = (roi.getArrayRegion(np.ones_like(data), self.imageitem)).T
    #                         print 'Cut:', cut.shape
    #
    #
    #                     if cut is not None:
    #                         if self.iscake:
    #
    #                             ma = np.ma.masked_array(data, mask=1 - (cut * self.dimg.cakemask))
    #                             chi = self.dimg.cakeqy
    #                             I = np.ma.average(ma, axis=1)
    #                             I = np.trim_zeros(I, 'b')
    #                             chi = chi[:len(I)]
    #                             I = np.trim_zeros(I, 'f')
    #                             chi = chi[-len(I):]
    #                             self.plotchiintegration([chi, I, [0, 255, 255]])
    #                         else:
    #
    #                             ai = config.activeExperiment.getAI().getPyFAI()
    #                             self.pool.apply_async(integration.chiintegratepyFAI,
    #                                                   args=(self.dimg.transformdata, self.dimg.mask, ai, self.iscake, cut,
    #                                                         [0, 255, 255]),
    #                                                   callback=self.chiintegrationrelay)
    #
    #
    #
    #                 else:
    #                     self.viewbox.removeItem(roi)
    #
    # def replotq(self):
    #     if not config.activeExperiment.iscalibrated:
    #         return None
    #
    #     cut = None
    #
    #     data = self.dimg.transformdata
    #
    #     ai = config.activeExperiment.getAI().getPyFAI()
    #     print 'centeroverride:', [c * config.activeExperiment.getvalue('Pixel Size X') for c in
    #                               self.getcenter()[::-1]]
    #     print self.getcenter()
    #
    #     xglobals.pool.apply_async(integration.radialintegratepyFAI,
    #                               args=((self.dimg.transformdata),
    #                                     (self.dimg.mask if not self.isremesh else self.dimg.remeshmask), ai, None, None,
    #                                     [c * config.activeExperiment.getvalue('Pixel Size X') for c in
    #                                      self.getcenter()[::-1]]),
    #                               callback=self.qintegrationrelay)
    #
    #     for roi in self.viewbox.addedItems:
    #         # try:
    #             if hasattr(roi, 'isdeleting'):
    #                 if not roi.isdeleting:
    #                     print type(roi)
    #                     cut = None
    #                     if issubclass(type(roi), pg.ROI) or issubclass(type(roi), pg.LinearRegionItem):
    #
    #                         cut = (roi.getArrayRegion(np.ones_like(data), self.imageitem)).T
    #                         print 'Cut:', cut.shape
    #
    #
    #
    #
    #
    #
    #
    #                     elif type(roi) is pg.LineSegmentROI:
    #
    #                         cut = self.region.getArrayRegion(data, self.imageitem)
    #
    #                         x = np.linspace(self.viewbox.mapSceneToView(self.region.getSceneHandlePositions(0)[1]).x(),
    #                                         self.viewbox.mapSceneToView(self.region.getSceneHandlePositions(1)[1]).x(),
    #                                         cut.__len__())
    #                         y = np.linspace(self.viewbox.mapSceneToView(self.region.getSceneHandlePositions(0)[1]).y(),
    #                                         self.viewbox.mapSceneToView(self.region.getSceneHandlePositions(1)[1]).y(),
    #                                         cut.__len__())
    #
    #                         q = pixel2q(x, y, self.dimg.experiment)
    #                         qmiddle = q.argmin()
    #                         leftq = -q[0:qmiddle]
    #                         rightq = q[qmiddle:]
    #
    #                         if leftq.__len__() > 1: self.plotwidget.qintegration.plot(leftq, cut[:qmiddle])
    #                         if rightq.__len__() > 1: self.plotwidget.qintegration.plot(rightq, cut[qmiddle:])
    #
    #
    #                     if cut is not None:
    #                         if self.iscake:
    #
    #                             ma = np.ma.masked_array(data, mask=1 - (cut * self.dimg.cakemask))
    #                             q = self.dimg.cakeqx / 10.
    #                             I = np.ma.average(ma, axis=0)
    #                             I = np.trim_zeros(I, 'b')
    #                             q = q[:len(I)]
    #                             I = np.trim_zeros(I, 'f')
    #                             q = q[-len(I):]
    #                             self.plotqintegration([q, I, [0, 255, 255]])
    #                         else:
    #
    #                             ai = config.activeExperiment.getAI().getPyFAI()
    #                             xglobals.pool.apply_async(integration.radialintegratepyFAI, args=(
    #                                 (self.dimg.transformdata),
    #                                 (self.dimg.mask), ai, cut, #if not self.isremesh else self.dimg.remeshmask
    #                                 [0, 255, 255], [c * config.activeExperiment.getvalue('Pixel Size X') for c in
    #                                                 self.getcenter()[::-1]]),
    #                                                       callback=self.qintegrationrelay)
    #
    #
    #                 else:
    #                     self.viewbox.removeItem(roi)
    #         # except Exception as ex:
    #         #     print 'Warning: error displaying ROI integration.'
    #         #     print ex.message
    #
    # def qintegrationrelay(self, *args, **kwargs):
    #     self.sigPlotQIntegration.emit(*args, **kwargs)
    #
    # def plotqintegration(self, result):
    #     (q, radialprofile, color) = result
    #     if color is None:
    #         color = [255, 255, 255]
    #     # cyan:[0, 255, 255]
    #     curve = self.plotwidget.qintegration.plot(q, radialprofile, pen=pg.mkPen(color=color))
    #     curve.setZValue(3 * 255 - sum(color))
    #     self.plotwidget.qintegration.update()
    #
    # def chiintegrationrelay(self, *args, **kwargs):
    #     print args
    #     self.sigPlotChiIntegration.emit(*args, **kwargs)
    #
    # def plotchiintegration(self, result):
    #     (chi, chiprofile, color) = result
    #     if color is None:
    #         color = [255, 255, 255]
    #     # cyan:[0, 255, 255]
    #     curve = self.plotwidget.chiintegration.plot(chi, chiprofile, pen=pg.mkPen(color=color))
    #     curve.setZValue(3 * 255 - sum(color))
    #     self.plotwidget.chiintegration.update()

    def polymask(self):
        maskroi = None
        for roi in self.viewbox.addedItems:
            if type(roi) is pg.PolyLineROI:
                maskroi = roi

        if maskroi is None:

            # Start with a box around the center
            left = config.activeExperiment.getvalue('Center X') - 100
            right = config.activeExperiment.getvalue('Center X') + 100
            up = config.activeExperiment.getvalue('Center Y') - 100
            down = config.activeExperiment.getvalue('Center Y') + 100

            # Add ROI item to the image
            self.maskROI = pg.PolyLineROI([[left, up], [left, down], [right, down], [right, up]], pen=(6, 9),
                                          closed=True)
            self.viewbox.addItem(self.maskROI)

            # Override the ROI's function to check if any points will be moved outside the boundary; False prevents move
            def checkPointMove(handle, pos, modifiers):
                p = self.viewbox.mapToView(pos)
                if 0 < p.y() < self.dimg.transformdata.shape[1] and 0 < p.x() < self.dimg.transformdata.shape[0]:
                    return True
                else:
                    return False

            self.maskROI.checkPointMove = checkPointMove

        else:  # If the mask is completed

            # Get the region of the image that was selected; unforunately the region is trimmed
            maskedarea = self.maskROI.getArrayRegion(np.ones_like(self.dimg.transformdata), self.imageitem)  # levels=(0, arr.max()
            # print maskedarea.shape

            # Decide how much to left and top pad based on the ROI bounding rectangle
            boundrect = self.viewbox.itemBoundingRect(self.maskROI)
            leftpad = max(boundrect.x(), 0)
            toppad = max(boundrect.y(), 0)


            # Pad the mask so it has the same shape as the image
            maskedarea = np.pad(maskedarea, ((int(leftpad), 0), (int(toppad), 0)), mode='constant')
            maskedarea = np.pad(maskedarea, (
                (0, self.dimg.transformdata.shape[0] - maskedarea.shape[0]), (0, self.dimg.transformdata.shape[1] - maskedarea.shape[1])),
                                mode='constant')

            # Add the masked area to the active mask
            config.activeExperiment.addtomask(1 - maskedarea)
            self.dimg.invalidatecache()

            # Draw the overlay
            # self.maskoverlay()

            # Remove the ROI
            self.viewbox.removeItem(self.maskROI)

            # Redraw the mask
            self.maskoverlay()

            # Redo the integration
            self.replot()


    def thresholdmask(self):
        threshold, ok = QtGui.QInputDialog.getInt(self, 'Threshold value','Input intensity threshold:',3,0,10000000)
        print 'threshold:',threshold

        if ok and threshold:
            kernelsize, ok = QtGui.QInputDialog.getInt(self, 'Neighborhood size', 'Neighborhood size (binary closing kernel size):', 2, 0,
                                                      10000000)
            if ok and kernelsize:
                mask = self.dimg.rawdata>threshold

                y,x = np.ogrid[-kernelsize:kernelsize+1,-kernelsize:kernelsize+1]
                kernel = x**2+y**2 <=kernelsize**2

                morphology.binary_closing(mask,kernel,output=mask) # write-back to mask

                config.activeExperiment.mask=mask
                self.maskoverlay()


    def maskoverlay(self):
        if self.iscake:
            mask = self.dimg.cakemask
        elif self.isremesh:
            mask = self.dimg.remeshmask
        else:
            mask = config.activeExperiment.mask

        if mask is None:
            self.maskimage.clear()
        else:
            # Draw the mask as a red channel image with an alpha mask
            msg.logMessage(('maskmax:', np.max(mask)),msg.DEBUG)
            invmask = 1 - mask
            self.maskimage.setImage(
                np.dstack((invmask, np.zeros_like(invmask), np.zeros_like(invmask), invmask)).astype(np.float),
                opacity=.5)
            self.maskimage.setLevels([0,1])

    # def finddetector(self):
    # # detector, mask = pipeline.loader.finddetector(self.imgdata)
    # if self.path is not None:
    # name, mask, detector = pipeline.loader.finddetectorbyfilename(self.path)
    #     else:
    #         name, mask, detector = pipeline.loader.finddetector(self.imgdata)
    #     if detector is not None:
    #         # name = detector.get_name()
    #         if mask is not None:
    #             self.experiment.addtomask(np.rot90(mask))
    #         self.experiment.setvalue('Pixel Size X', detector.pixel1)
    #         self.experiment.setvalue('Pixel Size Y', detector.pixel2)
    #         self.experiment.setvalue('Detector', name)
    #     return detector


    def exportimage(self):
        data = self.imageitem.image
        guesspath = self.paths[0]
        writer.writeimage(data, path=guesspath, headers=self.dimg.headers,dialog=True)

    def capture(self):
        captureroi = None
        for roi in self.viewbox.addedItems:
            if type(roi) is pg.RectROI:
                captureroi = roi

        if captureroi is None:

            # Add ROI item to the image
            self.captureROI = pg.RectROI(config.activeExperiment.center, (100, 100))
            self.viewbox.addItem(self.captureROI)

            # Override the ROI's function to check if any points will be moved outside the boundary; False prevents move
            def checkPointMove(handle, pos, modifiers):
                p = self.viewbox.mapToView(pos)
                if 0 < p.y() < self.dimg.data.shape[0] and 0 < p.x() < self.dimg.data.shape[1]:
                    return True
                else:
                    return False

            self.captureROI.checkPointMove = checkPointMove

        else:  # If the mask is completed

            # Get the shape
            lowerleft = [max(int(c), 0) for c in self.captureROI.pos()]
            topright = [max(int(s + p), 0) for s, p in zip(self.captureROI.size(), self.captureROI.pos())]

            dataregion = self.dimg.rawdata[lowerleft[0]:topright[0], lowerleft[1]:topright[1]]
            maskregion = self.dimg.mask[lowerleft[0]:topright[0], lowerleft[1]:topright[1]]
            guesspath = self.filepaths[0]

            qpar_min = self.getq(*lowerleft, mode='parallel') * 10
            qvrt_min = self.getq(*lowerleft, mode='z') * 10
            qpar_max = self.getq(*topright, mode='parallel') * 10
            qvrt_max = self.getq(*topright, mode='z') * 10

            headers = {'qpar_min': qpar_min, 'qpar_max': qpar_max, 'qvrt_min': qvrt_min, 'qvrt_max': qvrt_max}
            writer.writeimage(data=dataregion, mask=maskregion, headers=headers, guesspath=guesspath, dialog=True)

            # Remove the ROI
            self.viewbox.removeItem(self.captureROI)

    def clearsgoverlays(self):
        for item in self.viewbox.addedItems:
            from pipeline import spacegroups

            # print type(item), type(item) is spacegroups.peakoverlay
            if type(item) is spacegroups.peakoverlay:
                self.viewbox.removeItem(item)  # item is marked for removal, but must be hidden removal takes effect
                item.hide()

    def drawsgoverlay(self, peakoverlay):
        self.clearsgoverlays()

        peakoverlay.setCenter(*self.getcenter())

        #peakoverlay.setRect(QtCore.QRect(self.getcenter()[0],self.getcenter()[1],1,1))
        self.viewbox.addItem(peakoverlay)
        peakoverlay.enable(self.viewbox)


class timelineViewer(dimgViewer):
    sigAddTimelineData = QtCore.Signal(object,object)
    sigClearTimeline = QtCore.Signal()
    def __init__(self, simg=None, files=None, toolbar=None):
        self.variationcurve = dict()
        self.toolbar = toolbar

        if simg is None:
            simg = loader.loaddiffimage(files)

        self.simg = simg
        #dimg = simg.first()

        self.operationindex = 0

        super(timelineViewer, self).__init__(simg, toolbar=toolbar)

        #img = np.array(self.simg.thumbs)
        #img = (np.log(img * (img > 0) + (img < 1)))

        self.imgview.setImage(simg, xvals=simg.xvals(''))

        # self.imageitem.sigImageChanged.connect(self.setscale)

        # self.paths = dict(zip(range(len(paths)), sorted(paths)))
        # self.setvariationmode(0)
        # self.gotomax()
        # self.imgview.timeLine.sigPositionChanged.disconnect(self.imgview.timeLineChanged)

        self.highresimgitem = pg.ImageItem()

        self.viewbox.addItem(self.highresimgitem)
        self.highresimgitem.hide()

        # self.imgview.timeLine.sigPositionChangeFinished.connect(self.drawframeoverlay)
        # self.imgview.timeLine.sigDragged.connect(self.hideoverlay)
        # self.imgview.sigKeyRelease.connect(self.drawframeoverlay)

        # self.drawframeoverlay()
        self.imgview.autoRange()

        timelineplot = self.imgview.getRoiPlot()
        self.timeline = timelineplot.getPlotItem()
        self.timeline.showAxis('left', False)
        self.timeline.showAxis('bottom', False)
        self.timeline.showAxis('top', True)
        self.timeline.showGrid(x=True)

        self.timeline.getViewBox().setMouseEnabled(x=False, y=True)
        # self.timeline.setLabel('bottom', u'Frame #', '')

        # self.timeline.getViewBox().buildMenu()
        menu = self.timeline.getViewBox().menu
        operationcombo = QtGui.QComboBox()
        operationcombo.setObjectName('operationcombo')
        operationcombo.addItems(variationoperators.operations.keys())
        operationcombo.currentIndexChanged.connect(self.setvariationmode)
        opwidgetaction = QtGui.QWidgetAction(menu)
        opwidgetaction.setDefaultWidget(operationcombo)
        # need to connect it
        menu.addAction(opwidgetaction)

        # self.imgview.getHistogramWidget().item.setImageItem(self.highresimgitem)
        # self.imgview.getHistogramWidget().item.sigLevelChangeFinished.connect(self.updatelowresLUT)


        self.rescan()

    def redrawimage(self, returnimg=False):
        self.simg.cakemode = self.iscake
        self.simg.remeshmode = self.isremesh
        self.simg.radialsymmetrymode = self.isradialsymmetry
        self.simg.mirrorsymmetrymode = self.ismirrorsymmetry
        self.simg.logscale = self.islogintensity
        super(timelineViewer, self).redrawimage(returnimg)
        timelineplot = self.imgview.getRoiPlot()
        self.timeline = timelineplot.getPlotItem()
        self.timeline.getViewBox().setMouseEnabled(x=False, y=True)

    def processtimeline(self):
        self.toolbar.actionProcess.setChecked(False)
        self.rescan()

    def aborttimeline(self):
        pass

    def reduce(self):
        pass
        # self.skipframes = (self.variationy[0:-1] / self.variationy[1:]) > 0.1

    def appendimage(self, d, paths):  # WIP
        paths = [os.path.join(d, path) for path in paths]
        self.simg.appendimages(paths)



    def rescan(self):
        #return
        self.cleartimeline()
        variationoperators.experiment = config.activeExperiment


        # keys = range(len(self.simg))[:-1]# TODO: generalize!
        # values = list(variation.variationiterator(self.simg))[1:]
        # print 'keys:',keys
        # print 'values:',values
        # d = dict(zip(keys,values))
        # self.plotvariation(d)

        # Run on thread queue
        bg_variation = threads.iterator(callback_slot=self.sigAddTimelineData,
                                        finished_slot=self.processingfinished,
                                        parent=self)(variation.variationiterator)

        # xglobals.pool.apply_async(variation.scanvariation,args=(self.simg.filepaths),callback=self.testreceive)

        # variation = self.simg.scan(self.operationindex)
        # self.plotvariation(variation)

        for roi in self.viewbox.addedItems:
            # try:
            if hasattr(roi, 'isdeleting'):
                if not roi.isdeleting:
                    roi = roi.getArrayRegion(np.ones_like(self.imgview.imageItem.image), self.imageitem).T
                    # variation = self.simg.scan(self.operationindex, roi)
                    # self.plotvariation(variation, [0, 255, 255])
                    # Run on thread queue
                    # TODO: pull color from ROI, give ROIs deterministic colors with pyqtgraph.intColor
                    bg_variation(self.simg, self.operationindex, roi=roi)
                    break
                else:
                    self.viewbox.removeItem(roi)
                    # except Exception as ex:
                    # print 'Warning: error displaying ROI variation.'
                    #    print ex.message

        else:
            bg_variation(self.simg, self.operationindex)


    def processingfinished(self, *args, **kwargs):
        msg.showMessage('Processing complete.',4)

    def setvariationmode(self, index):
        self.operationindex = index

    def cleartimeline(self):
        self.sigClearTimeline.emit()

    # def plotvariation(self, variation, color=None):
    #     if len(variation) == 0:
    #         return None
    #
    #     if color is None:
    #         color = [255, 255, 255]
    #
    #     # TODO: plot variation with indices, and skipped frames; skip None's
    #
    #
    #
    #     t = np.array(variation.keys())
    #     d = np.array(variation.values())
    #
    #     # print variation
    #     d = d[t.argsort()]
    #     t = sorted(t)
    #
    #     self.timeline.enableAutoScale()
    #     # self.timeruler = TimeRuler(pen=pg.mkPen('#FFA500', width=3), movable=True)
    #
    #     print 'plottype:', type(d[0])
    #     if type(d[0]) in [float, int, np.float64]:
    #         self.timeline.plot(t, d, pen=pg.mkPen(color=color))
    #     elif type(d[0]) is np.ndarray:
    #         for dt in d:
    #             self.timeline.plot(dt[0], dt[1])

    # def redrawframe(self, index, time, forcelow=False):
    #     key = round(time)
    #     self.dimg = self.simg.getDiffImage(key)
    #     self.redrawimage(forcelow=forcelow)

    def gotomax(self):
        pass
        # self.parentwindow.timeruler.setValue(np.argmax(self.variationy))

    def replot(self):
        pass


class integrationwidget(QtGui.QTabWidget):

    def __init__(self, getViewer):
        super(integrationwidget, self).__init__()
        self.setTabPosition(self.West)

        self.getViewer = getViewer

        self.qintegration = qintegrationwidget()
        self.chiintegration = chiintegrationwidget()
        self.xintegration = xintegrationwidget()
        self.zintegration = zintegrationwidget()

        self.cakexintegration = cakexintegrationwidget()
        self.cakezintegration = cakezintegrationwidget()

        self.remeshqintegration = remeshqintegrationwidget()
        self.remeshchiintegration = remeshchiintegrationwidget()
        self.remeshxintegration = remeshxintegrationwidget()
        self.remeshzintegration = remeshzintegrationwidget()

        self.currentChanged.connect(self.replot)


    def movPosLine(self,*args,**kwargs):
        if self.currentIndex() == -1: return
        self.widget(self.currentIndex()).movPosLine(*args,**kwargs)

    def updatemodes(self):
        if not self.getViewer():
            return

        previouswidget = self.currentWidget()

        self.blockSignals(True)

        for i in range(self.count(),-1,-1):
            self.removeTab(i)

        if self.getViewer().iscake:
            self.addTab(self.cakexintegration, u'x')
            self.addTab(self.cakezintegration, u'z')
        elif self.getViewer().isremesh:
            self.addTab(self.remeshqintegration, u'q')
            self.addTab(self.remeshchiintegration, u'χ')
            self.addTab(self.remeshxintegration, u'x')
            self.addTab(self.remeshzintegration, u'z')
        else:
            self.addTab(self.qintegration, u'q')
            self.addTab(self.chiintegration, u'χ')
            self.addTab(self.xintegration, u'x')
            self.addTab(self.zintegration, u'z')

        newwidgets = [self.widget(i) for i in range(self.count())]
        if previouswidget in newwidgets: self.setCurrentWidget(previouswidget)

        self.blockSignals(False)



    def replot(self):
        viewer = self.getViewer()
        if not viewer: return
        self.updatemodes()
        dimg = viewer.dimg
        rois = viewer.getROIs()
        imageitem = viewer.imageitem
        self.widget(self.currentIndex()).replot(dimg,rois,imageitem)

    def hidePosLine(self):
        if self.currentIndex() > -1:
            self.widget(self.currentIndex()).posLine.hide()


class integrationsubwidget(pg.PlotWidget):
    integrationfunction = None
    iscake = False
    isremesh = False

    def __init__(self,axislabel):
        super(integrationsubwidget, self).__init__()
        self.setLabel('bottom', axislabel, '')
        self.posLine = pg.InfiniteLine(angle=90, movable=False, pen=pg.mkPen('#FFA500'))
        self.posLine.setVisible(False)
        self.addItem(self.posLine)

        self.iscleared = True
        self.requestkey = 0
        self.hasrun=False


    def replot(self, dimg, rois, imageitem):
        data = dimg.transformdata
        mask = dimg.transformmask
        if self.integrationfunction is None:
            raise NotImplementedError
        try:
            self.applyintegration(self.integrationfunction,dimg,rois,data,mask,imageitem)
        except ValueError:
            msg.logMessage('Maybe the roi was too far away?',msg.DEBUG)

    def applyintegration(self,integrationfunction,dimg,rois,data,mask,imageitem):
        self.requestkey += 1
        self.iscleared = False

        if self.iscake:
            qvrt = dimg.cakeqy
            qpar = dimg.cakeqx
        elif self.isremesh:
            qvrt = dimg.remeshqy
            qpar = dimg.remeshqx
        else:
            qvrt = None
            qpar = None

        # xglobals.pool.apply_async(integrationfunction, args=(data, mask, dimg.experiment.getAI().getPyFAI(), None,
        #                                                                None, self.requestkey, qvrt, qpar),
        #                           callback=self.replotcallback)


        # threads.method(callback_slot=self.plotresult)(self.test)('test')
        args = data, mask, dimg.experiment.getAI().getPyFAI(), None, None, self.requestkey, qvrt, qpar


        # replot roi integration
        for roi in rois:
            if roi.isdeleting:
                roi.deleteLater()
                continue

            cut = (roi.getArrayRegion(np.ones_like(dimg.transformdata), imageitem)).T
            msg.logMessage(('Cut:', cut.shape),msg.DEBUG)

            if cut is not None:
                # xglobals.pool.apply_async(integrationfunction,
                #                           args=(data, mask, dimg.experiment.getAI().getPyFAI(), cut, [0, 255, 255], self.requestkey, qvrt, qpar),
                #                           callback=self.replotcallback)
                args=data, mask, dimg.experiment.getAI().getPyFAI(), cut, [0, 255, 255], self.requestkey, qvrt, qpar

        threads.method(callback_slot=self.plotresult)(integrationfunction)(*args)

    def movPosLine(self, q,qx,qz,dimg=None):
        pass #raise NotImplementedError

    def plotresult(self, x, y, color, requestkey): #x, y, color, requestkey
        # print 'RESULT:',x, y, color, requestkey #x, y, color, requestkey, QtCore.QThread.currentThread()

        if requestkey == self.requestkey:
            if not self.iscleared:
                self.plotItem.clear()
                self.addItem(self.posLine)
                self.iscleared = True
            if color is None:
                color = [255, 255, 255]
            y[y<=0]=1.E-9
            curve = self.plotItem.plot(np.array(x), np.nan_to_num(np.array(y).astype(float)), pen=pg.mkPen(color=color))
            curve.setZValue(3 * 255 - sum(color))

            self.plotItem.update()


class qintegrationwidget(integrationsubwidget):
    integrationfunction = staticmethod(integration.qintegrate)

    def __init__(self):
        super(qintegrationwidget, self).__init__(axislabel=u'q (\u212B\u207B\u00B9)')

    def movPosLine(self,q,qx,qz,dimg=None):
        self.posLine.setPos(q)
        self.posLine.show()


class chiintegrationwidget(integrationsubwidget):
    integrationfunction = staticmethod(integration.chiintegratepyFAI)

    def __init__(self):
        super(chiintegrationwidget, self).__init__(axislabel=u'χ (Degrees)')

    def movPosLine(self,q, qx, qz, dimg=None):
        self.posLine.setPos(np.rad2deg(np.arctan2(qz,qx)))
        self.posLine.show()

class xintegrationwidget(integrationsubwidget):
    integrationfunction = staticmethod(integration.xintegrate)

    def __init__(self):
        super(xintegrationwidget, self).__init__(axislabel=u'q<sub>x</sub> (\u212B\u207B\u00B9)')

    def movPosLine(self,q, qx, qz, dimg=None):
        self.posLine.setPos(qx)
        self.posLine.show()

class zintegrationwidget(integrationsubwidget):
    integrationfunction = staticmethod(integration.zintegrate)

    def __init__(self):
        super(zintegrationwidget, self).__init__(axislabel=u'q<sub>z</sub> (\u212B\u207B\u00B9)')

    def movPosLine(self,q, qx, qz, dimg=None):
        self.posLine.setPos(qz)
        self.posLine.show()

class cakexintegrationwidget(integrationsubwidget):
    iscake = True
    integrationfunction = staticmethod(integration.cakexintegrate)

    def __init__(self):
        super(cakexintegrationwidget, self).__init__(axislabel=u'χ (Degrees)')

    def movPosLine(self,q, qx, qz, dimg=None):
        self.posLine.setPos(np.rad2deg(np.arctan2(qx,qz)))
        self.posLine.show()

class cakezintegrationwidget(integrationsubwidget):

    iscake = True
    integrationfunction = staticmethod(integration.cakezintegrate)

    def __init__(self):
        super(cakezintegrationwidget, self).__init__(axislabel=u'q (\u212B\u207B\u00B9)')

    def movPosLine(self,q, qx, qz, dimg=None):
        self.posLine.setPos(q)
        self.posLine.show()

class remeshqintegrationwidget(integrationsubwidget):

    isremesh=True
    integrationfunction = staticmethod(integration.remeshqintegrate)

    def __init__(self):
        super(remeshqintegrationwidget, self).__init__(axislabel=u'q (\u212B\u207B\u00B9)')

    def movPosLine(self,q,qx,qz,dimg=None):
        self.posLine.setPos(q)
        self.posLine.show()

class remeshchiintegrationwidget(integrationsubwidget):

    isremesh=True
    integrationfunction = staticmethod(integration.remeshchiintegrate)

    def __init__(self):
        super(remeshchiintegrationwidget, self).__init__(axislabel=u'χ (Degrees)')

    def movPosLine(self,q, qx, qz, dimg=None):
        self.posLine.setPos(np.rad2deg(np.arctan2(qz, qx)))
        self.posLine.show()

class remeshxintegrationwidget(integrationsubwidget):

    isremesh=True
    integrationfunction = staticmethod(integration.remeshxintegrate)

    def __init__(self):
        super(remeshxintegrationwidget, self).__init__(axislabel=u'q (\u212B\u207B\u00B9)')

    def movPosLine(self,q, qx,qz,dimg=None):
        self.posLine.setPos(qx)
        self.posLine.show()

class remeshzintegrationwidget(integrationsubwidget):

    isremesh=True
    integrationfunction = staticmethod(integration.remeshzintegrate)

    def __init__(self):
        super(remeshzintegrationwidget, self).__init__(axislabel=u'q (\u212B\u207B\u00B9)')

    def movPosLine(self,q, qx,qz,dimg=None):
        self.posLine.setPos(qz)
        self.posLine.show()


def getHistogram(self, bins='auto', step='auto', targetImageSize=None, targetHistogramSize=500, **kwds):
    """Returns x and y arrays containing the histogram values for the current image.
    For an explanation of the return format, see numpy.histogram().

    The *step* argument causes pixels to be skipped when computing the histogram to save time.
    If *step* is 'auto', then a step is chosen such that the analyzed data has
    dimensions roughly *targetImageSize* for each axis.

    The *bins* argument and any extra keyword arguments are passed to
    np.histogram(). If *bins* is 'auto', then a bin number is automatically
    chosen based on the image characteristics:

    * Integer images will have approximately *targetHistogramSize* bins,
      with each bin having an integer width.
    * All other types will have *targetHistogramSize* bins.

    This method is also used when automatically computing levels.
    """
    if self.image is None:
        return None, None

    if not targetImageSize: targetImageSize = min(200, self.image.shape[0], self.image.shape[1])

    if step == 'auto':
        step = (np.ceil(self.image.shape[0] / targetImageSize),
                np.ceil(self.image.shape[1] / targetImageSize))
    if np.isscalar(step):
        step = (step, step)
    stepData = self.image[::int(step[0]), ::int(step[1])]

    if bins == 'auto':
        if stepData.dtype.kind in "ui":
            mn = stepData.min()
            mx = stepData.max()
            step = np.ceil((mx - mn) / 500.)
            bins = np.arange(mn, mx + 1.01 * step, step, dtype=np.int)
            if len(bins) == 0:
                bins = [mn, mx]
        else:
            bins = 500

    kwds['bins'] = bins
    hist = np.histogram(stepData, **kwds)
    hist[0][0] = 0
    return hist[1][:-1], hist[0]


pg.ImageItem.getHistogram = getHistogram

class ImageView(pg.ImageView):
    sigKeyRelease = QtCore.Signal()
    def __init__(self,*args,**kwargs):
        self.actionLog_Intensity=kwargs['actionLog_Intensity']
        del kwargs['actionLog_Intensity']
        super(ImageView, self).__init__(*args,**kwargs)

    def buildMenu(self):
        super(ImageView, self).buildMenu()
        self.menu.removeAction(self.normAction)

    def timeIndex(self, slider):
        ## Return the time and frame index indicated by a slider
        if self.image is None:
            return (0, 0)

        t = slider.value()

        xv = self.tVals
        if xv is None:
            ind = int(t)
        else:
            if len(xv) < 2:
                return (0, 0)
            inds = np.argwhere(xv <= t)  # <- The = is import to reach the last value
            if len(inds) < 1:
                return (0, t)
            ind = inds[-1, 0]
        return ind, t

    def setImage(self,*args,**kwargs):
        super(ImageView, self).setImage(*args,**kwargs)
        # if self.actionLog_Intensity.isChecked():
        #     levelmin = np.log(self.levelMin)
        #     levelmax = np.log(self.levelMax)
        #     if np.isnan(levelmin): levelmin = 0
        #     if np.isnan(levelmax): levelmax = 1
        #     if np.isinf(levelmin): levelmin = 0
        #     msg.logMessage(('min:',levelmin),msg.DEBUG)
        #     msg.logMessage(('max:',levelmax),msg.DEBUG)
        #
        #     self.ui.histogram.setLevels(levelmin, levelmax)

    # def updateImage(self, autoHistogramRange=True): # inject logarithm action
    #     ## Redraw image on screen
    #     if self.image is None:
    #         return
    #
    #     image = self.getProcessedImage()
    #
    #     if autoHistogramRange:
    #         self.ui.histogram.setHistogramRange(self.levelMin, self.levelMax)
    #     if self.axes['t'] is None:
    #         self.imageItem.updateImage(np.log(image * (image> 0) + (image < 1)) if self.actionLog_Intensity.isChecked() else image)
    #     else:
    #         self.ui.roiPlot.show()
    #         self.imageItem.updateImage(np.log(image[self.currentIndex] * (image[self.currentIndex]> 0) + (image[self.currentIndex] < 1)) if self.actionLog_Intensity.isChecked() else image[self.currentIndex])


from scipy.signal import fftconvolve


class fxsviewer(ImageView):
    def __init__(self, paths=None, **kwargs):
        super(fxsviewer, self).__init__()
        if paths is None:
            paths = []

        conv = None
        avg = None
        for path in paths:
            dimg = loader.diffimage(filepath=path)
            if avg is None: avg = np.zeros((100, 100))

            def autocorrelate(A):
                return fftconvolve(A, A[::-1])

            conv = np.apply_along_axis(autocorrelate, axis=0, arr=dimg.cake)
            avg += conv[99:, :] + conv[:100, :]

        self.setImage(avg)


class xasviewer(pg.PlotWidget):
    def __init__(self, paths, *args, **kwargs):
        super(xasviewer, self).__init__(*args, **kwargs)
        for path in paths:
            for i in range(2, 16):
                self.plot(loader.loadxfs(path)[:, 2], loader.loadxfs(path)[:, i], pen=(i, 16))


class pluginModeWidget(QtGui.QWidget):
    def __init__(self, plugins):
        super(pluginModeWidget, self).__init__()
        self.setLayout(QtGui.QHBoxLayout())

        self.font = QtGui.QFont()
        self.font.setPointSize(16)
        self.plugins = plugins

        self.reload()

    def reload(self):
        w = self.layout().takeAt(0)
        while w:
            w.widget().deleteLater()
            del w
            w = self.layout().takeAt(0)

        for key, plugin in self.plugins.items():
            if plugin.enabled:
                if plugin.instance.hidden:
                    continue

                button = QtGui.QPushButton(plugin.name)
                button.setFlat(True)
                button.setFont(self.font)
                button.setProperty('isMode', True)
                button.setAutoFillBackground(False)
                button.setCheckable(True)
                button.setAutoExclusive(True)
                button.clicked.connect(plugin.activate)
                if plugin is self.plugins.values()[0]:
                    button.setChecked(True)
                self.layout().addWidget(button)
                label = QtGui.QLabel('|')
                label.setFont(self.font)
                label.setStyleSheet('background-color:#111111;')
                self.layout().addWidget(label)

        self.layout().takeAt(self.layout().count() - 1).widget().deleteLater()  # Delete the last pipe symbol


class previewwidget(pg.GraphicsLayoutWidget):
    """
    top-left preview
    """

    def __init__(self):
        super(previewwidget, self).__init__()
        self.view = self.addViewBox(lockAspect=True, enableMenu=False)
        self.imageitem = pg.ImageItem()
        self.textitem = pg.TextItem()
        self.imgdata = None
        self.setMinimumHeight(250)

        self.view.addItem(self.imageitem)
        self.view.addItem(self.textitem)
        self.textitem.hide()
        self.setSizePolicy(QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Minimum)

        # def textItemBounds(axis, frac=1.0, orthoRange=None):
        #     b = self.textitem.boundingRect()
        #     sx, sy = self.view.viewPixelSize()
        #     x, y = sx*b.width(), sy*b.height()
        #     if axis == 0: return (-x/2, x/2)
        #     if axis == 1: return (0, y)
        #
        # self.textitem.dataBounds = textItemBounds

    def loaditem(self, item):
        if isinstance(item, str) or isinstance(item, unicode):
            if os.path.isfile(item) or os.path.isdir(item):
                item = loader.loadimage(item)
            else:
                self.setText(item)
                return
        try:
            self.setImage(item)
        except TypeError:
            self.imageitem.clear()

    def setImage(self, imgdata):
        self.imageitem.clear()
        self.textitem.hide()
        self.imgdata = imgdata
        self.imageitem.setImage(np.rot90(np.log(self.imgdata * (self.imgdata > 0) + (self.imgdata < 1)), 3),
                                autoLevels=True)

    def setText(self, text):
        self.textitem.setText(text)
        self.textitem.setFont(QtGui.QFont('Zero Threes'))  # Not working for some reason
        self.imageitem.clear()
        self.textitem.show()


class fileTreeWidget(QtGui.QTreeView):
    sigOpenFiles = QtCore.Signal(list)
    sigOpenDir = QtCore.Signal(str)

    def __init__(self):
        super(fileTreeWidget, self).__init__()
        self.filetreemodel = QtGui.QFileSystemModel()
        self.setModel(self.filetreemodel)
        self.filetreepath = pathtools.getRoot()
        self.treerefresh(self.filetreepath)
        header = self.header()
        self.setHeaderHidden(True)
        for i in range(1, 4):
            header.hideSection(i)
        filefilter = ['*' + ext for ext in
                      loader.acceptableexts]  # ["*.tif", "*.edf", "*.fits", "*.nxs", "*.hdf", "*.cbf"]
        self.filetreemodel.setNameFilters(filefilter)
        self.filetreemodel.setNameFilterDisables(False)
        self.filetreemodel.setResolveSymlinks(True)
        self.expandAll()
        self.sortByColumn(0, QtCore.Qt.AscendingOrder)
        self.setSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Expanding)
        self.setSelectionMode(self.ExtendedSelection)
        self.setIconSize(QtCore.QSize(16, 16))

        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.contextMenu)

        self.doubleClicked.connect(self.doubleclickevent)

    def contextMenu(self, position):
        menu = QtGui.QMenu()

        actionOpen = QtGui.QAction('Open', self)
        actionOpen.triggered.connect(self.openActionTriggered)
        menu.addAction(actionOpen)

        menu.exec_(self.viewport().mapToGlobal(position))

    def openActionTriggered(self):
        indices = self.selectedIndexes()
        paths = [self.filetreemodel.filePath(index) for index in indices]

        self.sigOpenFiles.emit(paths)

    def doubleclickevent(self, index):
        path = self.filetreemodel.filePath(index)

        if os.path.isfile(path):
            self.sigOpenFiles.emit([path])
        elif os.path.isdir(path):
            self.sigOpenDir.emit(path)
        else:
            msg.logMessage(('Error on index (what is that?):', index),msg.ERROR)

    def treerefresh(self, path=None):
        """
        Refresh the file tree, or switch directories and refresh
        """
        if path is None:
            path = self.filetreepath

        root = QtCore.QDir(path)
        self.filetreemodel.setRootPath(root.absolutePath())
        self.setRootIndex(self.filetreemodel.index(root.absolutePath()))
        self.show()


class frameproptable(pg.TableWidget):
    """
    Widget for displaying hierarchical python data structures
    (eg, nested dicts, lists, and arrays), adapted from pyqtgraph datatree.
    """

    def __init__(self, parent=None, data=None):
        super(frameproptable, self).__init__(parent)
        self.setVerticalScrollMode(self.ScrollPerPixel)
        self.setData(data)
        self.setColumnCount(2)
        self.setHorizontalHeaderLabels(['Parameter', 'Value'])
        self.horizontalHeader().setStretchLastSection(True)
        self.verticalHeader().hide()
        self.setSizePolicy(QtGui.QSizePolicy.MinimumExpanding,QtGui.QSizePolicy.MinimumExpanding)
        self.setSelectionMode(QtGui.QAbstractItemView.SingleSelection)
        self.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)

        useAsMenu = QtGui.QMenu(u'Use as...',parent=self.contextMenu)
        useAsMenu.addAction('Beam Energy').triggered.connect(self.useAsEnergy)
        useAsMenu.addAction('Downstream Intensity').triggered.connect(self.useAsI1)
        useAsMenu.addAction('Timeline Axis').triggered.connect(self.useAsTimeline)
        self.contextMenu.addMenu(useAsMenu)

    def setData(self,data):
        if data is None:
            self.setVisible(False)
            return
        data=sorted(data.items())
        self.setHidden(len(data) == 0)
        super(frameproptable, self).setData(data)

    def sizeHint(self):
        return QtCore.QSize(self.parent().width(),self.parent().height()/2)

    def useAsI1(self):
        config.activeExperiment.setHeaderMap('I1 AI',self.getSelectedKey())

    def useAsEnergy(self):
        config.activeExperiment.setHeaderMap('Beam Energy',self.getSelectedKey())

    def useAsTimeline(self):
        config.activeExperiment.setHeaderMap('Timeline Axis', self.getSelectedKey())

    def getSelectedKey(self):
        return self.item(self.selectedIndexes()[0].row(),0).value


def pixel2q(x, y, experiment):
    # SWITCH TO PYFAI GEOMETRY

    if x is None:
        x = experiment.getvalue('Center X')
    if y is None:
        y = experiment.getvalue('Center Y')

    r = np.sqrt((x - experiment.getvalue('Center X')) ** 2 + (y - experiment.getvalue('Center Y')) ** 2)
    theta = np.arctan2(r * experiment.getvalue('Pixel Size X'),
                       experiment.getvalue('Detector Distance'))
    wavelength = experiment.getvalue('Wavelength')
    return 4 * np.pi / wavelength * np.sin(theta / 2) * 1e-10


class filesListWidget(QtGui.QWidget):
    def __init__(self, *args, **kwargs):
        super(filesListWidget, self).__init__()
        self.horizontalLayout = QtGui.QHBoxLayout(self)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.paths = QtGui.QListWidget(self)
        self.paths.setSelectionMode(QtGui.QAbstractItemView.MultiSelection)
        self.paths.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)
        self.paths.setObjectName("paths")
        self.horizontalLayout.addWidget(self.paths)
        self.verticalLayout_8 = QtGui.QVBoxLayout()
        self.verticalLayout_8.setObjectName("verticalLayout_8")
        self.addfilesbutton = QtGui.QToolButton(self)
        self.addfilesbutton.setObjectName("addfiles")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("xicam/gui/icons_43.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.addfilesbutton.setIcon(icon)
        self.verticalLayout_8.addWidget(self.addfilesbutton)
        self.removefilesbutton = QtGui.QToolButton(self)
        self.removefilesbutton.setObjectName("removefiles")
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap("xicam/gui/icons_58.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.removefilesbutton.setIcon(icon2)
        self.verticalLayout_8.addWidget(self.removefilesbutton)
        spacerItem = QtGui.QSpacerItem(20, 40, QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Expanding)
        self.verticalLayout_8.addItem(spacerItem)
        self.horizontalLayout.addLayout(self.verticalLayout_8)
        self.addfilesbutton.clicked.connect(self.addfiles)
        self.removefilesbutton.clicked.connect(self.removefiles)

    def addfiles(self):
        paths, ok = QtGui.QFileDialog.getOpenFileNames(None, 'Add files to Batch', os.curdir,
                                                       '*' + ' *'.join(loader.acceptableexts))
        self.paths.addItems(paths)

    def removefiles(self):
        for index in self.paths.selectedIndexes():
            item = self.paths.takeItem(index.row())
            item = None



#! /usr/bin/env python

import os
import h5py
import numpy as np
import pipeline
import viewer
from pipeline import loader
from PySide import QtCore, QtGui
from xicam import xglobals


class SPOT_H5DSet:
    def __init__(self, dat, txt):
        self.data = dat
        tmp = dat.attrs
        self.dim1 = tmp['dim2']
        self.dim2 = tmp['dim3']
        self.date = tmp['date']
        self.header = tmp['edfHeaderInBytes']
        self.text = txt

    @classmethod
    def calibration(cls, grp):
        if not isinstance(grp, h5py._hl.group.Group):
            raise TypeError('argument must of a h5py Group')

        keys = grp.keys()
        if not len(keys) == 2:
            raise ValueError('Wrong h5py Group passed')

        txt = None
        edf = None
        for key in keys:
            if str(key).endswith('edf'):
                edf = grp[key]
            if str(key).endswith('txt'):
                txt = grp[key]
            return cls(edf, txt)


class SPOT_H5Node:
    def __init__(self, grp):
        if not isinstance(grp, h5py._hl.group.Group):
            raise TypeError('argument must of a h5py Group')
        keys = grp.keys()
        if not len(keys) == 3:
            raise ValueError('Wrong h5py Group passed')

        txt = None
        edf = None
        for key in keys:
            if key == unicode('calibration'):
                self.calibration = SPOT_H5DSet.calibration(grp[key])
            if str(key).endswith('.edf'):
                edf = grp[key]
            if str(key).endswith('.txt'):
                txt = grp[key]
        if edf == None or txt == None:
            raise IOError('file not found')
        self.data = SPOT_H5DSet(edf, txt)


class SPOT_H5:
    def __init__(self, filename):
        if not os.path.splitext(filename)[1] == '.h5':
            raise IOError('File extension must be .h5')

        self.h5 = h5py.File(filename)
        self.metadata = self.h5.attrs
        self.isTiled = False
        node = self.h5.values()[0].values()[0]
        keys = node.keys()

        if len(keys) == 2 and unicode('high') in keys:
            self.isTiled = True
            self.high = SPOT_H5Node(node['high'])
            self.node = SPOT_H5Node(node['low'])
        elif len(keys) == 3 and unicode('calibration') in keys:
            self.node = SPOT_H5Node(node)

    def getCalibrationImg(self):
        data = self.node.calibration.data
        shape = data.shape[1:]
        return data.value.reshape(shape)

    def getCalibrationHeader(self):
        return self.node.calibration.header

    def getImage(self):
        data = self.node.data.data
        shape = data.shape[1:]
        return data.value.reshape(shape)

    def getHeader(self):
        return self.node.data.header

    def getHighImage(self):
        data = self.high.data.data
        shape = data.shape[1:]
        return data.value.reshape(shape)

    def getHighHeader(self):
        return self.high.data.header

    def getText(self):
        return self.node.data.text.value[0]

    def getCalibrationText(self):
        return self.node.calibration.text.value[0]


class SPOTViewerPlugin(viewer.ViewerPlugin):
    name = "SPOTH5"

    def openfiles(self, filenames):
        h5 = SPOT_H5(filenames[0])
        data = h5.getCalibrationImg()
        self.opendata(data=data)
        self.calibrate()



        # stich high and low
        if h5.isTiled:
            low_header = {'Detector Vertical': 0, 'Detector Horizontal': 0}
            high_header = {'Detector Vertical': 6.88, 'Detector Horizontal': 0}
            low_data = h5.getImage()
            high_data = h5.getHighImage()
            data = loader.loadstitched(None, None, low_data, high_data, low_header, high_header)
            self.opendata(data)


if __name__ == '__main__':
    foo = SPOT_H5('/Users/dkumar/Downloads/20151004_234826_PorRun5_283.h5')

    txt = foo.getText()
    print txt
    dat = foo.getCalibrationImg()
    # saveedf("AgB_000.edf", dat, hdr)
    



from pyFAI import azimuthalIntegrator
import numpy

import pyFAI

import logging
logger = logging.getLogger("pyFAI.azimuthalIntegrator")

from pyFAI import version_info

if version_info.minor>=16 and version_info.major==0:
    class AzimuthalIntegrator(azimuthalIntegrator.AzimuthalIntegrator):
        USE_LEGACY_MASK_NORMALIZATION = False
else:

    # monkey patch to correct auto-inversion of masks
    class AzimuthalIntegrator(azimuthalIntegrator.AzimuthalIntegrator):


        def create_mask(self, data, mask=None,
                        dummy=None, delta_dummy=None, mode="normal"):
            """
            Combines various masks into another one.

            @param data: input array of data
            @type data: ndarray
            @param mask: input mask (if none, self.mask is used)
            @type mask: ndarray
            @param dummy: value of dead pixels
            @type dummy: float
            @param delta_dumy: precision of dummy pixels
            @type delta_dummy: float
            @param mode: can be "normal" or "numpy" (inverted) or "where" applied to the mask
            @type mode: str

            @return: the new mask
            @rtype: ndarray of bool

            This method combine two masks (dynamic mask from *data &
            dummy* and *mask*) to generate a new one with the 'or' binary
            operation.  One can adjust the level, with the *dummy* and
            the *delta_dummy* parameter, when you consider the *data*
            values needs to be masked out.

            This method can work in two different *mode*:

                * "normal": False for valid pixels, True for bad pixels
                * "numpy": True for valid pixels, false for others

            This method tries to accomodate various types of masks (like
            valid=0 & masked=-1, ...) and guesses if an input mask needs
            to be inverted.
            """
            shape = data.shape
            #       ^^^^   this is why data is mandatory !
            if mask is None:
                mask = self.mask
            if mask is None:
                mask = numpy.zeros(shape, dtype=bool)
            elif mask.min() < 0 and mask.max() == 0:  # 0 is valid, <0 is invalid
                mask = (mask < 0)
            else:
                mask = mask.astype(bool)
            # if mask.sum(dtype=int) > mask.size // 2:                              # TERRRIBLEE!
            #     logger.warning("Mask likely to be inverted as more"
            #                    " than half pixel are masked !!!")
            #     numpy.logical_not(mask, mask)
            if (mask.shape != shape):
                try:
                    mask = mask[:shape[0], :shape[1]]
                except Exception as error:  # IGNORE:W0703
                    logger.error("Mask provided has wrong shape:"
                                 " expected: %s, got %s, error: %s" %
                                 (shape, mask.shape, error))
                    mask = numpy.zeros(shape, dtype=bool)
            if dummy is not None:
                if delta_dummy is None:
                    numpy.logical_or(mask, (data == dummy), mask)
                else:
                    numpy.logical_or(mask,
                                     abs(data - dummy) <= delta_dummy,
                                     mask)
            if mode == "numpy":
                numpy.logical_not(mask, mask)
            elif mode == "where":
                mask = numpy.where(numpy.logical_not(mask))
            return mask

azimuthalIntegrator.__dict__['AzimuthalIntegrator']=AzimuthalIntegrator


import numpy as np
import loader
import scipy.ndimage
import warnings
import variationoperators
import msg


def variationiterator(simg,operationindex,roi=None):
    for i in range(len(simg)):
        yield simg.calcVariation(i, operationindex, roi)

def scanvariation(filepaths):
    simg = loader.multifilediffimage2(filepaths)
    for t in range(len(simg)):
        variationoperators.chisquared(simg,t,None)

def filevariation(operationindex, filea, c, filec, roi=None):
    p = loader.loadimage(filea)
    # c, _ = loader.loadpath(fileb)
    n = loader.loadimage(filec)
    # print 'previous frame:' + filea
    # print p
    return variation(operationindex, p, c, n, roi)

# Deprecating...
def variation(operationindex, imga, imgb=None, imgc=None, roi=None):
    if imga is not None and imgb is not None and imgc is not None:
        try:
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                # TODO: REFACTOR THIS TO USE THUMBS!!!!!
                p = scipy.ndimage.zoom(imga, 0.1, order=1)
                c = scipy.ndimage.zoom(imgb, 0.1, order=1)
                n = scipy.ndimage.zoom(imgc, 0.1, order=1)
                p = scipy.ndimage.gaussian_filter(p, 3)
                c = scipy.ndimage.gaussian_filter(c, 3)
                n = scipy.ndimage.gaussian_filter(n, 3)
                if roi is not None:
                    roi = scipy.ndimage.zoom(roi, 0.1, order=1)
                    roi = np.flipud(roi)
                else:
                    roi = 1
            with np.errstate(divide='ignore'):
                return variationoperators.operations.values()[operationindex](p, c, n, roi, None, None)
        except TypeError:
            msg.logMessage('Variation could not be determined for a frame.',msg.ERROR)
    else:
        msg.logMessage('Variation could not be determined for a frame.',msg.ERROR)
    return 0